# -*- coding: utf-8 -*-
#
# Copyright (c) 2018-2022, Geoffrey M. Poore
# All rights reserved.
#
# Licensed under the BSD 3-Clause License:
# http://opensource.org/licenses/BSD-3-Clause
#


from __future__ import annotations


import collections
import hashlib
import json
import os
import pathlib
import platform
import re
import subprocess
import sys
import tempfile
import textwrap
from typing import List, Optional, Sequence, Tuple, Union
from ..code_chunks import CodeChunk, Include
from .. import err
from .. import message
from .. import util
from ..codebraid_defaults import CodebraidDefaults
from .base import Converter


# Pandoc node types mapped to integers representing the layout of node
# contents (if any).
# https://github.com/jgm/pandocfilters/blob/master/pandocfilters.py
pandoc_block_nodes = {
    'Plain': 1,
    'Para': 1,
    'CodeBlock': 2,
    'RawBlock': 2,
    'BlockQuote': 1,
    'OrderedList': 2,
    'BulletList': 1,
    'DefinitionList': 1,
    'Header': 3,
    'HorizontalRule': 0,
    'Table': 5,
    'Div': 2,
    'Null': 0,
}
pandoc_inline_nodes = {
    'Str': 1,
    'Emph': 1,
    'Strong': 1,
    'Strikeout': 1,
    'Superscript': 1,
    'Subscript': 1,
    'SmallCaps': 1,
    'Quoted': 2,
    'Cite': 2,
    'Code': 2,
    'Space': 0,
    'LineBreak': 0,
    'Math': 2,
    'RawInline': 2,
    'Link': 3,
    'Image': 3,
    'Note': 1,
    'SoftBreak': 0,
    'Span': 2,
}
pandoc_nodes = {**pandoc_block_nodes, **pandoc_inline_nodes}
pandoc_code_nodes = {k: v for k, v in pandoc_nodes.items() if k.startswith('Code')}
pandoc_raw_nodes = {k: v for k, v in pandoc_nodes.items() if k.startswith('Raw')}
pandoc_code_raw_nodes = {**pandoc_code_nodes, **pandoc_raw_nodes}




class PandocError(err.CodebraidError):
    pass




def _get_class_processors():
    '''
    Create dict mapping Pandoc classes to processing functions.
    '''
    def class_lang_or_unknown(code_chunk, options, class_index, class_name):
        if 'lang' in options:
            if class_name.startswith('cb.') or class_name.startswith('cb-'):
                code_chunk.errors.append(message.SourceError('Unknown or unsupported Codebraid command "{0}"'.format(class_name)))
            else:
                code_chunk.errors.append(message.SourceError('Unknown non-Codebraid class'))
        elif class_index > 0:
            if class_name.startswith('cb.') or class_name.startswith('cb-'):
                code_chunk.errors.append(message.SourceError('Unknown or unsupported Codebraid command "{0}"'.format(class_name)))
            else:
                code_chunk.errors.append(message.SourceError('The language/format "{0}" must be the first class for code chunk'.format(class_name)))
        else:
            options['lang'] = class_name

    def class_codebraid_command(code_chunk, options, class_index, class_name):
        if 'codebraid_command' in options:
            code_chunk.errors.append(message.SourceError('Only one Codebraid command can be applied per code chunk'))
        else:
            if '.' in class_name:
                options['codebraid_command'] = class_name.split('.', 1)[1]
            elif '-' in class_name:
                options['codebraid_command'] = class_name.split('-', 1)[1]
            else:
                raise ValueError

    def class_line_anchors(code_chunk, options, class_index, class_name):
        if 'line_anchors' in options:
            code_chunk.errors.append(message.CanExecSourceError('Duplicate line anchor class for code chunk'))
        else:
            options['line_anchors'] = True

    def class_line_numbers(code_chunk, options, class_index, class_name):
        if 'line_numbers' in options:
            code_chunk.errors.append(message.CanExecSourceError('Duplicate line numbering class for code chunk'))
        else:
            options['line_numbers'] = True

    codebraid_commands = {f'cb.{k}': class_codebraid_command for k in CodeChunk.commands}
    codebraid_commands.update({f'cb-{k}': class_codebraid_command for k in CodeChunk.commands})
    line_anchors = {k: class_line_anchors for k in ('lineAnchors', 'line-anchors', 'line_anchors')}
    line_numbers = {k: class_line_numbers for k in ('line_numbers', 'numberLines', 'number-lines', 'number_lines')}
    return collections.defaultdict(lambda: class_lang_or_unknown,
                                   {**codebraid_commands,
                                    **line_anchors,
                                    **line_numbers})


def _get_keyval_processors():
    '''
    Create dict mapping Pandoc key-value attributes to processing functions.
    '''
    # Options like `include` have sub-options, which need to be translated
    # into a dict.  In a markup language with more expressive option syntax,
    # this would be supported as `include={key=value, ...}`.  For Pandoc
    # Markdown, the alternatives are keys like `include.file` or
    # `include_file`.  In either case, the base option name `include`
    # functions as a namespace and needs to be protected from being
    # overwritten by an invalid option.  Track those namespaces with a set.
    namespace_keywords = set()

    def keyval_generic(code_chunk, options, key, value):
        if key in options:
            code_chunk.errors.append(message.SourceError('Duplicate "{0}" attribute for code chunk'.format(key)))
        elif key in namespace_keywords:
            code_chunk.errors.append(message.SourceError('Invalid key "{0}" for code chunk'.format(key)))
        else:
            options[key] = value

    def keyval_bool(code_chunk, options, key, value):
        if key in options:
            code_chunk.errors.append(message.SourceError('Duplicate "{0}" attribute for code chunk'.format(key)))
        elif value not in ('true', 'false'):
            code_chunk.errors.append(message.SourceError('Attribute "{0}" must be true or false for code chunk'.format(key)))
        else:
            options[key] = value == 'true'

    def keyval_int(code_chunk, options, key, value):
        if key in options:
            code_chunk.errors.append(message.SourceError('Duplicate "{0}" attribute for code chunk'.format(key)))
        else:
            try:
                value = int(value)
            except (ValueError, TypeError):
                code_chunk.errors.append(message.SourceError('Attribute "{0}" must be integer for code chunk'.format(key)))
            else:
                options[key] = value

    def keyval_first_number(code_chunk, options, key, value):
        if 'first_number' in options:
            code_chunk.errors.append(message.CanExecSourceError('Duplicate first line number attribute for code chunk'))
        else:
            try:
                value = int(value)
            except (ValueError, TypeError):
                pass
            options['first_number'] = value

    def keyval_line_anchors(code_chunk, options, key, value):
        if 'line_anchors' in options:
            code_chunk.errors.append(message.CanExecSourceError('Duplicate line anchor attribute for code chunk'))
        elif value not in ('true', 'false'):
            code_chunk.errors.append(message.CanExecSourceError('Attribute "{0}" must be true or false for code chunk'.format(key)))
        else:
            options['line_anchors'] = value == 'true'

    def keyval_line_numbers(code_chunk, options, key, value):
        if 'line_numbers' in options:
            code_chunk.errors.append(message.CanExecSourceError('Duplicate line numbering attribute for code chunk'))
        elif value not in ('true', 'false'):
            code_chunk.errors.append(message.CanExecSourceError('Attribute "{0}" must be true or false for code chunk'.format(key)))
        else:
            options['line_numbers'] = value == 'true'

    def keyval_namespace(code_chunk, options, key, value):
        if '.' in key:
            namespace, sub_key = key.split('.', 1)
        else:
            namespace, sub_key = key.split('_', 1)
        if namespace in options:
            sub_options = options[namespace]
        else:
            sub_options = {}
            options[namespace] = sub_options
        if sub_key in sub_options:
            code_chunk.errors.append(message.SourceError('Duplicate "{0}" attribute for code chunk'.format(key)))
        else:
            sub_options[sub_key] = value

    expand_tabs = {k: keyval_bool for k in ['{0}_expand_tabs'.format(dsp) if dsp else 'expand_tabs'
                                            for dsp in ('', 'markup', 'copied_markup', 'code', 'stdout', 'stderr')]}
    first_number = {k: keyval_first_number for k in ('first_number', 'startFrom', 'start-from', 'start_from')}
    include = {'include_{0}'.format(k): keyval_namespace for k in Include.keywords}
    line_anchors = {k: keyval_line_anchors for k in ('lineAnchors', 'line-anchors', 'line_anchors')}
    line_numbers = {k: keyval_line_numbers for k in ('line_numbers', 'numberLines', 'number-lines', 'number_lines')}
    rewrap_lines = {k: keyval_bool for k in ['{0}_rewrap_lines'.format(dsp) if dsp else 'rewrap_lines'
                                             for dsp in ('', 'markup', 'copied_markup', 'code', 'stdout', 'stderr')]}
    rewrap_width = {k: keyval_int for k in ['{0}_rewrap_width'.format(dsp) if dsp else 'rewrap_width'
                                            for dsp in ('', 'markup', 'copied_markup', 'code', 'stdout', 'stderr')]}
    tab_size = {k: keyval_int for k in ['{0}_tab_size'.format(dsp) if dsp else 'tab_size'
                                         for dsp in ('', 'markup', 'copied_markup', 'code', 'stdout', 'stderr')]}
    namespace_keywords.add('include')
    return collections.defaultdict(lambda: keyval_generic,
                                   {'complete': keyval_bool,
                                    'copy': keyval_generic,
                                    'example': keyval_bool,
                                    **expand_tabs,
                                    **first_number,
                                    **include,
                                    'jupyter_timeout': keyval_int,
                                    **line_anchors,
                                    **line_numbers,
                                    'live_output': keyval_bool,
                                    'name': keyval_generic,
                                    'outside_main': keyval_bool,
                                    **rewrap_lines,
                                    **rewrap_width,
                                    **tab_size})




class PandocCodeChunk(CodeChunk):
    '''
    Code chunk for Pandoc Markdown.
    '''
    def __init__(self,
                 node: dict,
                 parent_node: dict,
                 parent_node_list: list,
                 parent_node_list_index: int,
                 origin_name: Optional[str]=None,
                 origin_start_line_number: Optional[int]=None):
        super().__pre_init__()

        self.node = node
        self.parent_node = parent_node
        self.parent_node_list = parent_node_list
        self.parent_node_list_index = parent_node_list_index

        node_id, node_classes, node_kvpairs = node['c'][0]
        self.node_id = node_id if not node_id.startswith('codebraid-sourcepos') else ''
        self.node_classes = node_classes
        self.node_kvpairs = node_kvpairs
        self.node_sourcepos_id = node_id
        self.node_sourcepos_classes = []
        self.node_sourcepos_kvpairs = []
        code = node['c'][1]
        options = {}

        # Preprocess options
        inline = node['t'] == 'Code'
        for n, c in enumerate(node_classes):
            if c.startswith('codebraid-sourcepos'):
                self.node_sourcepos_classes.append(c)
                continue
            self._class_processors[c](self, options, n, c)
        if self.node_sourcepos_classes:
            self.node_classes = [c for c in self.node_classes if not c.startswith('codebraid-sourcepos')]
        for k, v in node_kvpairs:
            if k.startswith('codebraid-sourcepos'):
                self.node_sourcepos_kvpairs.append([k, v])
                continue
            self._kv_processors[k](self, options, k, v)
        if self.node_sourcepos_kvpairs:
            self.node_kvpairs = [[k, v] for (k, v) in self.node_kvpairs if not k.startswith('codebraid-sourcepos')]
        # All processed data from classes and key-value pairs is stored in
        # `options`, but only some of these are valid Codebraid options.
        # Remove those that are not and store in temp variables.
        codebraid_command = options.pop('codebraid_command', None)
        line_anchors = options.pop('lineAnchors', None)

        # Process options
        super().__init__(codebraid_command, code, options, origin_name=origin_name,
                         origin_start_line_number=origin_start_line_number, inline=inline)

        # Work with processed options -- now use `self.options`
        if inline and self.options['example']:
            if parent_node['t'] not in ('Plain', 'Para') or 'codebraid_pseudonode' in parent_node or len(parent_node_list) > 1:
                self.source_errors.insert(0, 'Option "example" is only allowed for inline code that is in a paragraph by itself')
                self.options['example'] = False
        pandoc_id = ''  # Could use node_id, but then must check duplication
        pandoc_classes = []
        pandoc_kvpairs = []
        lang = options.get('lang', None)
        if lang is not None:
            if lang.endswith('_repl'):
                lang = lang.rsplit('_')[0]
            pandoc_classes.append(lang)
            if self.command == 'repl':
                pandoc_classes.append('repl')
        if line_anchors:
            pandoc_classes.append('lineAnchors')
        if self.options.get('code_line_numbers', False):
            pandoc_classes.append('numberLines')
        # Can't handle `startFrom` yet here, because if it is `next`, then
        # the value depends on which other code chunks end up in the session.
        # Starting line number is determined when output is generated.

        self.pandoc_id = pandoc_id
        self.pandoc_classes = pandoc_classes
        self.pandoc_kvpairs = pandoc_kvpairs
        self._output_nodes = None
        self._as_markup_lines = None
        self._as_example_markup_lines = None


    _class_processors = _get_class_processors()
    _kv_processors = _get_keyval_processors()

    # This may need additional refinement in future depending on allowed values
    _unquoted_kv_value_re = re.compile(r'[A-Za-z$_+\-][A-Za-z0-9$_+\-:]*')


    def finalize_line_numbers(self, code_start_line_number):
        super().finalize_line_numbers(code_start_line_number)
        first_number = self.options['code_first_number']
        if first_number == 'next':
            first_number = str(self.code_start_line_number)
        else:
            first_number = str(first_number)
        self.pandoc_kvpairs.append(['startFrom', first_number])


    @property
    def attr_hash(self):
        attr_list = []
        attr_list.append('{')
        if self.node_id:
            attr_list.append(f'#{self.node_id}')
        attr_list.extend(f'.{c}' for c in self.node_classes)
        for k, v in self.node_kvpairs:
            v = v.replace('\\', '\\\\').replace('"', '\\"')
            attr_list.append(f'{k}="{v}"')
        attr_list.append('}')
        attr_str = ' '.join(attr_list)
        hasher = hashlib.sha1()
        hasher.update(attr_str.encode('utf8'))
        return hasher.hexdigest()

    @property
    def code_hash(self):
        hasher = hashlib.sha1()
        if self.placeholder_code_lines:
            hasher.update(self.placeholder_code_str.encode('utf8'))
        else:
            hasher.update(self.code_str.encode('utf8'))
        return hasher.hexdigest()


    def finalize_after_copy(self):
        if self.options['lang'] is None:
            self.pandoc_classes.insert(0, self.copy_chunks[0].options['lang'])
        self.options.finalize_after_copy()


    @property
    def output_nodes(self):
        '''
        A list of nodes representing output, or representing source errors if
        those prevented execution.
        '''
        if self._output_nodes is not None:
            return self._output_nodes
        nodes = []
        if self.index == 0 and self.code_collection.errors:
            msgs_list = []
            error_class_list = ['error']
            for error in self.code_collection.errors:
                if error.is_refed:
                    continue
                if msgs_list:
                    msgs_list.append('')
                msgs_list.append(f'{error.display_name.upper()} in "{self.origin_name}" near line {self.origin_start_line_number}:')
                msgs_list.extend(error.message)
                error_class = error.type[0].lower() + error.type[1:]
                if error_class not in error_class_list:
                    error_class_list.append(error_class)
            if msgs_list:
                if self.inline:
                    nodes.append({'t': 'Code', 'c': [['', error_class_list, []], ' '.join(msgs_list)]})
                else:
                    nodes.append({'t': 'CodeBlock', 'c': [['', error_class_list, []], '\n'.join(msgs_list)]})
                return nodes
        if self.errors.has_non_stderr:
            msgs_list = []
            error_class_list = ['error']
            for error in self.errors:
                if error.is_stderr:
                    continue
                if msgs_list:
                    msgs_list.append('')
                msgs_list.append(f'{error.display_name.upper()} in "{self.origin_name}" near line {self.origin_start_line_number}:')
                msgs_list.extend(error.message)
                error_class = error.type[0].lower() + error.type[1:]
                if error_class not in error_class_list:
                    error_class_list.append(error_class)
            if self.inline:
                nodes.append({'t': 'Code', 'c': [['', error_class_list, []], ' '.join(msgs_list)]})
            else:
                nodes.append({'t': 'CodeBlock', 'c': [['', error_class_list, []], '\n'.join(msgs_list)]})
            return nodes
        t_code = 'Code' if self.inline else 'CodeBlock'
        t_raw = 'RawInline' if self.inline else 'RawBlock'
        unformatted_nodes = []
        for output, format in self.options['show'].items():
            if output in ('markup', 'copied_markup'):
                unformatted_nodes.append({'t': t_code, 'c': [['', ['markdown'], []], self.layout_output(output, format)]})
            elif output == 'code':
                unformatted_nodes.append({'t': t_code, 'c': [[self.pandoc_id, self.pandoc_classes, self.pandoc_kvpairs], self.layout_output(output, format)]})
            elif output == 'repl':
                if not self.repl_lines:
                    continue
                unformatted_nodes.append({'t': t_code, 'c': [[self.pandoc_id, self.pandoc_classes, self.pandoc_kvpairs], self.layout_output(output, format)]})
            elif output in ('expr', 'stdout', 'stderr'):
                if output == 'stderr' and self.errors.has_stderr:
                    stderr_output_lines = self.stderr_lines.copy()
                    for error in self.errors:
                        if error.is_ref and error.message:
                            stderr_output_lines.extend(error.message)
                    if format == 'verbatim':
                        if stderr_output_lines:
                            unformatted_nodes.append({'t': t_code, 'c': [['', ['stderr', 'error'], []], self.layout_output(output, format, stderr_output_lines)]})
                    elif format == 'verbatim_or_empty':
                        unformatted_nodes.append({'t': t_code, 'c': [['', ['stderr', 'error'], []], self.layout_output(output, format, stderr_output_lines)]})
                    elif format == 'raw':
                        if getattr(self, output+'_lines'):
                            unformatted_nodes.append({'t': t_raw, 'c': ['markdown', self.layout_output(output, format, stderr_output_lines)]})
                    else:
                        raise ValueError
                    continue
                if format == 'verbatim':
                    if getattr(self, output+'_lines'):
                        unformatted_nodes.append({'t': t_code, 'c': [['', [output], []], self.layout_output(output, format)]})
                elif format == 'verbatim_or_empty':
                    unformatted_nodes.append({'t': t_code, 'c': [['', [output], []], self.layout_output(output, format)]})
                elif format == 'raw':
                    if getattr(self, output+'_lines'):
                        unformatted_nodes.append({'t': t_raw, 'c': ['markdown', self.layout_output(output, format)]})
                else:
                    raise ValueError
            elif output == 'rich_output':
                if not self.rich_output:
                    continue
                for ro in self.rich_output:
                    ro_data = ro['data']
                    ro_files = ro['files']
                    for fmt_and_text_display in format:
                        if ':' in fmt_and_text_display:
                            fmt, fmt_text_display = fmt_and_text_display.split(':')
                        else:
                            fmt = fmt_and_text_display
                            fmt_text_display = self.options.rich_text_default_display.get(fmt)
                        fmt_mime_type = self.options.mime_map[fmt]
                        if fmt_mime_type not in ro_data:
                            continue
                        if fmt_mime_type in ro_files:
                            image_node = {'t': 'Image', 'c': [['', ['richOutput'], []], [], [ro_files[fmt_mime_type], '']]}
                            if self.inline:
                                unformatted_nodes.append(image_node)
                            else:
                                para_node = {'t': 'Para', 'c': [image_node]}
                                unformatted_nodes.append(para_node)
                            break
                        data = ro_data[fmt_mime_type]
                        lines = util.splitlines_lf(data)
                        if fmt in ('latex', 'html', 'markdown'):
                            if fmt == 'latex':
                                raw_fmt = 'tex'
                            else:
                                raw_fmt = fmt
                            if fmt_text_display == 'raw':
                                unformatted_nodes.append({'t': t_raw, 'c': [raw_fmt, self.layout_output(output, 'raw', lines)]})
                            elif fmt_text_display == 'verbatim':
                                if lines:
                                    unformatted_nodes.append({'t': t_code, 'c': [['', [fmt], []], self.layout_output(output, 'verbatim', lines)]})
                            elif fmt_text_display == 'verbatim_or_empty':
                                unformatted_nodes.append({'t': t_code, 'c': [['', [fmt], []], self.layout_output(output, 'verbatim', lines)]})
                            else:
                                raise ValueError
                            break
                        if fmt == 'plain':
                            if fmt_text_display == 'raw':
                                unformatted_nodes.append({'t': t_raw, 'c': ['markdown', self.layout_output(output, fmt_text_display, lines)]})
                            elif fmt_text_display == 'verbatim':
                                if lines:
                                    unformatted_nodes.append({'t': t_code, 'c': [['', [], []], self.layout_output(output, 'verbatim', lines)]})
                            elif fmt_text_display == 'verbatim_or_empty':
                                unformatted_nodes.append({'t': t_code, 'c': [['', [], []], self.layout_output(output, 'verbatim', lines)]})
                            else:
                                raise ValueError
                            break
                        raise ValueError
            else:
                raise ValueError
        if unformatted_nodes:
            # Prevent adjacent nodes from merging unintentionally when
            # converted through intermediate Markdown
            for node, next_node in zip(unformatted_nodes[:-1], unformatted_nodes[1:]):
                nodes.append(node)
                if node['t'] == t_raw or next_node['t'] == t_raw:
                    if self.inline:
                        nodes.append({'t': 'Str', 'c': ' '})
                    else:
                        nodes.append({'t': 'Para', 'c': []})
                elif self.inline and node['t'] == next_node['t'] == t_code:
                    nodes.append({'t': 'Str', 'c': ' '})
            nodes.append(unformatted_nodes[-1])
        self._output_nodes = nodes
        return nodes


    @property
    def as_markup_lines(self):
        if self._as_markup_lines is not None:
            return self._as_markup_lines
        self._as_markup_lines = self._generate_markdown_lines()
        return self._as_markup_lines

    @property
    def as_example_markup_lines(self):
        if self._as_example_markup_lines is not None:
            return self._as_example_markup_lines
        self._as_example_markup_lines = self._generate_markdown_lines(example=True)
        return self._as_example_markup_lines

    def _generate_markdown_lines(self, example=False):
        '''
        Generate an approximation of the Markdown source that created the
        node.  This is used with `show=markup` and in creating examples that
        show Markdown source plus output.
        '''
        attr_list = []
        if self.node_id:
            attr_list.append('#{0}'.format(self.node_id))
        for c in self.node_classes:
            attr_list.append('.{0}'.format(c))
        hide_keys = set(self.options.get('hide_markup_keys', []))
        if example:
            hide_keys.add('example')
        for k, v in self.node_kvpairs:
            if k not in hide_keys:
                # Valid keys don't need quoting, some values may
                if not self._unquoted_kv_value_re.match(v):
                    v = '"{0}"'.format(v.replace('\\', '\\\\').replace('"', '\\"'))
                attr_list.append('{0}={1}'.format(k, v))
        if self.placeholder_code_lines:
            code_lines = self.placeholder_code_lines
            code_str = code_lines[0]
        else:
            code_lines = self.code_lines
            code_str = self.code_str
        if self.inline:
            code_strip = code_str.strip(' ')
            if code_strip.startswith('`'):
                code_str = ' ' + code_str
            if code_strip.endswith('`'):
                code_str = code_str + ' '
            delim = '`'
            while delim in code_str:
                delim += '`'
            md_lines = ['{delim}{code}{delim}{{{attr}}}'.format(delim=delim, code=code_str, attr=' '.join(attr_list))]
        elif not self.placeholder_code_lines or code_str:
            delim = '```'
            while delim in code_str:
                delim += '```'
            md_lines = ['{delim}{{{attr}}}'.format(delim=delim, attr=' '.join(attr_list)), *code_lines, delim]
        else:
            md_lines = ['```{{{attr}}}'.format(attr=' '.join(attr_list)), '```']
        return md_lines


    def only_code_output(self, format):
        if format != 'codebraid_preview':
            raise ValueError
        output_list = []
        if self.index == 0 and self.code_collection.errors:
            msgs_list = []
            error_class_list = ['error']
            for error in self.code_collection.errors:
                if error.is_refed:
                    continue
                if msgs_list:
                    msgs_list.append('')
                msgs_list.append(f'{error.display_name.upper()} in "{self.origin_name}" near line {self.origin_start_line_number}:')
                msgs_list.extend(error.message)
                error_class = error.type[0].lower() + error.type[1:]
                if error_class not in error_class_list:
                    error_class_list.append(error_class)
            if msgs_list:
                if self.inline:
                    output_list.append(self._as_markdown_inline_code(' '.join(msgs_list), classes=error_class_list))
                else:
                    output_list.append(self._as_markdown_block_code('\n'.join(msgs_list), classes=error_class_list))
                return output_list
        if self.errors.has_non_stderr:
            msgs_list = []
            error_class_list = ['error']
            for error in self.errors:
                if error.is_stderr:
                    continue
                if msgs_list:
                    msgs_list.append('')
                msgs_list.append(f'{error.display_name.upper()} in "{self.origin_name}" near line {self.origin_start_line_number}:')
                msgs_list.extend(error.message)
                error_class = error.type[0].lower() + error.type[1:]
                if error_class not in error_class_list:
                    error_class_list.append(error_class)
            if self.inline:
                output_list.append(self._as_markdown_inline_code(' '.join(msgs_list), classes=error_class_list))
            else:
                output_list.append(self._as_markdown_block_code('\n'.join(msgs_list), classes=error_class_list))
            return output_list
        for output, format in self.options['show'].items():
            if output in ('markup', 'copied_markup'):
                output_list.append(self._as_markdown_code(self.layout_output(output, format), inline=self.inline, classes=['markdown']))
            elif output == 'code':
                output_list.append(self._as_markdown_code(self.layout_output(output, format), inline=self.inline,
                                                          id=self.pandoc_id, classes=self.pandoc_classes, keyval=self.pandoc_kvpairs))
            elif output == 'repl':
                if not self.repl_lines:
                    continue
                output_list.append(self._as_markdown_code(self.layout_output(output, format), inline=self.inline,
                                                          id=self.pandoc_id, classes=self.pandoc_classes, keyval=self.pandoc_kvpairs))
            elif output in ('expr', 'stdout', 'stderr'):
                if output == 'stderr' and self.errors.has_stderr:
                    stderr_output_lines = self.stderr_lines.copy()
                    for error in self.errors:
                        if error.is_ref and error.message:
                            stderr_output_lines.extend(error.message)
                    if format == 'verbatim':
                        if stderr_output_lines:
                            output_list.append(self._as_markdown_code(self.layout_output(output, format, stderr_output_lines),
                                               inline=self.inline, classes=['stderr', 'error']))
                    elif format == 'verbatim_or_empty':
                        output_list.append(self._as_markdown_code(self.layout_output(output, format, stderr_output_lines),
                                           inline=self.inline, classes=['stderr', 'error']))
                    elif format == 'raw':
                        if getattr(self, output+'_lines'):
                            output_list.append(self._as_markdown_included_text(self.layout_output(output, format, stderr_output_lines), inline=self.inline))
                    else:
                        raise ValueError
                    continue
                if format == 'verbatim':
                    if getattr(self, output+'_lines'):
                        output_list.append(self._as_markdown_code(self.layout_output(output, format), inline=self.inline, classes=[output]))
                elif format == 'verbatim_or_empty':
                    output_list.append(self._as_markdown_code(self.layout_output(output, format), inline=self.inline, classes=[output]))
                elif format == 'raw':
                    if getattr(self, output+'_lines'):
                        output_list.append(self._as_markdown_included_text(self.layout_output(output, format), inline=self.inline))
                else:
                    raise ValueError
            elif output == 'rich_output':
                if not self.rich_output:
                    continue
                for ro in self.rich_output:
                    ro_data = ro['data']
                    ro_files = ro['files']
                    for fmt_and_text_display in format:
                        if ':' in fmt_and_text_display:
                            fmt, fmt_text_display = fmt_and_text_display.split(':')
                        else:
                            fmt = fmt_and_text_display
                            fmt_text_display = self.options.rich_text_default_display.get(fmt)
                        fmt_mime_type = self.options.mime_map[fmt]
                        if fmt_mime_type not in ro_data:
                            continue
                        if fmt_mime_type in ro_files:
                            output_list.append(self._as_markdown_image(ro_files[fmt_mime_type], classes=['richOutput']))
                            break
                        data = ro_data[fmt_mime_type]
                        lines = util.splitlines_lf(data)
                        if fmt in ('latex', 'html', 'markdown'):
                            if fmt == 'latex':
                                raw_fmt = 'tex'
                            else:
                                raw_fmt = fmt
                            if fmt_text_display == 'raw':
                                output_list.append(self._as_markdown_included_text(self.layout_output(output, 'raw', lines), inline=self.inline))
                            elif fmt_text_display == 'verbatim':
                                if lines:
                                    output_list.append(self._as_markdown_code(self.layout_output(output, 'verbatim', lines), inline=self.inline, classes=[fmt]))
                            elif fmt_text_display == 'verbatim_or_empty':
                                output_list.append(self._as_markdown_code(self.layout_output(output, 'verbatim', lines), inline=self.inline, classes=[fmt]))
                            else:
                                raise ValueError
                            break
                        if fmt == 'plain':
                            if fmt_text_display == 'raw':
                                output_list.append(self._as_markdown_included_text(self.layout_output(output, fmt_text_display, lines), inline=self.inline))
                            elif fmt_text_display == 'verbatim':
                                if lines:
                                    output_list.append(self._as_markdown_code(self.layout_output(output, 'verbatim', lines), inline=self.inline))
                            elif fmt_text_display == 'verbatim_or_empty':
                                output_list.append(self._as_markdown_code(self.layout_output(output, 'verbatim', lines), inline=self.inline))
                            else:
                                raise ValueError
                            break
                        raise ValueError
            else:
                raise ValueError
        return output_list


    def _as_markdown_attr(self,
                          id: Optional[str]=None,
                          classes: Optional[list[str]]=None,
                          keyval: Optional[list]=None,
                          raw: Optional[str]=None):
        if raw:
            if id or classes or keyval:
                raise TypeError
            return f'{{={raw}}}'
        attr_list = []
        if id or classes or keyval or raw:
            attr_list.append('{ ')
            if id:
                attr_list.append(f'#{id} ')
            if classes:
                attr_list.extend(f'.{c} ' for c in classes)
            if keyval:
                attr_list.extend('{0}="{1}" '.format(k, v.replace('"', '\\"')) for k, v in keyval)
            attr_list.append('}')
        return ''.join(attr_list)

    def _as_markdown_inline_code(self, text: str, *, id=None, classes=None, keyval=None, raw=None, protect_start=True):
        md_list = []
        if protect_start:
            md_list.append('[]{.codebraid-protect-inline=true}')
        delim = '`'
        while delim in text:
            delim += '`'
        md_list.append(delim)
        if text.startswith(' ') and text.endswith(' '):  # https://spec.commonmark.org/0.30/#code-spans
            md_list.append(f'{delim} {text} {delim}')
        else:
            if text.startswith('`'):
                md_list.append(f'{delim} {text}')
            else:
                md_list.append(f'{delim}{text}')
            if text.endswith('`'):
                md_list.append(f' {delim}')
            else:
                md_list.append(delim)
        md_list.append(self._as_markdown_attr(id, classes, keyval, raw))
        return ''.join(md_list)

    def _as_markdown_block_code(self, text: str, *, id=None, classes=None, keyval=None, raw=None):
        md_list = []
        delim = '```'
        while delim in text:
            delim += '```'
        md_list.append(f'{delim}{self._as_markdown_attr(id, classes, keyval, raw)}\n')
        md_list.append(text)
        if not text.endswith('\n'):
            md_list.append('\n')
        md_list.append(delim)
        return ''.join(md_list)

    def _as_markdown_code(self, text: str, *, inline: Optional[bool], id=None, classes=None, keyval=None, raw=None, protect_start=True):
        if inline:
            return self._as_markdown_inline_code(text, id=id, classes=classes, keyval=keyval, raw=raw, protect_start=protect_start)
        else:
            return self._as_markdown_block_code(text, id=id, classes=classes, keyval=keyval, raw=raw)

    def _as_markdown_included_text(self, text: str, *, inline: Optional[bool], protect_start=True):
        if inline and protect_start:
            return f'[]{{.codebraid-protect-inline-start}}{text}'
        return text

    def _as_markdown_image(self, path: str, *, alt='', id=None, classes=None, keyval=None):
        return f'![{alt}]({path}){self._as_markdown_attr(id, classes, keyval)}'


    def update_parent_node(self):
        '''
        Update parent node with output.
        '''
        if not self.options['example']:
            index = self.parent_node_list_index
            if self.inline or (not self.node_sourcepos_classes and not self.node_sourcepos_kvpairs):
                output_nodes = self.output_nodes
            else:
                output_nodes =  [{'t': 'Div', 'c': [[self.node_sourcepos_id, self.node_sourcepos_classes, self.node_sourcepos_kvpairs], self.output_nodes]}]
            self.parent_node_list[index:index+1] = output_nodes
            return
        markup_node = {'t': 'CodeBlock', 'c': [['', [], []], self.layout_output('example_markup', 'verbatim')]}
        example_div_contents = [{'t': 'Div', 'c': [['', ['exampleMarkup'], []], [markup_node]]}]
        output_nodes = self.output_nodes
        if output_nodes:
            if self.inline:
                # `output_nodes` are all inline, but will be inserted into a
                # div, so need a block-level wrapper
                output_nodes = [{'t': 'Para', 'c': output_nodes}]
            example_div_contents.append({'t': 'Div', 'c': [['', ['exampleOutput'], []], output_nodes]})
        if self.inline or (not self.node_sourcepos_classes and not self.node_sourcepos_kvpairs):
            example_div_node = {'t': 'Div', 'c': [['', ['example'], []], example_div_contents]}
        else:
            example_div_node = {'t': 'Div', 'c': [[self.node_sourcepos_id, ['example', *self.node_sourcepos_classes], self.node_sourcepos_kvpairs], example_div_contents]}
        if self.inline:
            # An inline node can't be replaced with a div directly.  Instead,
            # its block-level parent node is redefined.  The validity of this
            # operation is checked during initialization.
            parent_para_plain_node = self.parent_node
            parent_para_plain_node['t'] = 'Div'
            parent_para_plain_node['c'] = example_div_node['c']
        else:
            index = self.parent_node_list_index
            self.parent_node_list[index] = example_div_node




def _get_walk_closure(enumerate=enumerate, isinstance=isinstance, list=list, dict=dict):
    def walk_node_list(node_list, parent_node, type_filter=None, skip_note_contents=False, in_note=False):
        '''
        Walk all AST nodes in a list, recursively descending to walk all child
        nodes as well.  The walk function is written so that it is only ever
        called on lists, reducing recursion depth and reducing the number of
        times the walk function is called.  Thus, it is never called on `Str`
        nodes and other leaf nodes, which will typically make up the vast
        majority of nodes.

        DefinitionLists are handled specially to wrap terms in fake Plain
        nodes, which are marked so that they can be identified later if
        necessary.  This simplifies processing.

        Yields a tuple containing a node, its parent node, parent list,
        parent list index, and a boolean indicating whether the node is inside
        a note.

        If `type_filter` is provided, it must be an object like a `set()` that
        supports membership checks via `in`.  Only nodes with types in
        `type_filter` will be yielded.

        If `skip_notes` is true, recursion skips nodes inside notes.
        '''
        for index, obj in enumerate(node_list):
            if isinstance(obj, dict):
                try:
                    node_type = obj['t']
                except KeyError:
                    continue
                if type_filter is None or node_type in type_filter:
                    yield (obj, parent_node, node_list, index, in_note)
                if 'c' in obj:
                    obj_contents = obj['c']
                    if isinstance(obj_contents, list):
                        if node_type == 'Note':
                            if skip_note_contents:
                                continue
                            in_note = True
                            yield from walk_node_list(obj_contents, obj, type_filter, skip_note_contents, in_note)
                        elif node_type != 'DefinitionList':
                            yield from walk_node_list(obj_contents, obj, type_filter, skip_note_contents, in_note)
                        else:
                            for elem in obj_contents:
                                term, definition = elem
                                pseudonode = {'t': 'Plain', 'c': term, 'codebraid_pseudonode': True}
                                if type_filter is None or 'Plain' in type_filter:
                                    yield (pseudonode, obj, elem, 0, in_note)
                                yield from walk_node_list(term, pseudonode, type_filter, skip_note_contents, in_note)
                                yield from walk_node_list(definition, obj, type_filter, skip_note_contents, in_note)
            elif isinstance(obj, list):
                yield from walk_node_list(obj, parent_node, type_filter, skip_note_contents, in_note)
    return walk_node_list
walk_node_list = _get_walk_closure()




class PandocConverter(Converter):
    '''
    Converter based on Pandoc (https://pandoc.org/).

    Pandoc is used to parse input into a JSON-based AST.  Code nodes in the
    AST are located, processed, and then replaced.  Next, the AST is converted
    back into the original input format (or possibly another format), and
    reparsed into a new AST.  This allows raw code output to be interpreted as
    markup.  Finally, the new AST can be converted into the output format.
    '''
    def __init__(self, *,
                 pandoc_path: Optional[Union[str, pathlib.Path]]=None,
                 pandoc_file_scope: Optional[bool]=False,
                 other_pandoc_args_at_load: Optional[list]=None,
                 from_format: Optional[str]=None,
                 scroll_sync: bool=False,
                 **kwargs):
        if from_format is not None and not isinstance(from_format, str):
            raise TypeError
        from_format, from_format_pandoc_extensions = self._split_format_extensions(from_format)

        super().__init__(from_format=from_format, **kwargs)

        if pandoc_path is None:
            pandoc_path = pathlib.Path('pandoc')
        else:
            pandoc_path = pathlib.Path(pandoc_path)
            if self.expandvars:
                pandoc_path = pathlib.Path(os.path.expandvars(pandoc_path))
            if self.expanduser:
                pandoc_path = pandoc_path.expanduser()
        try:
            proc = subprocess.run([str(pandoc_path), '--version'], stdout=subprocess.PIPE, stderr=subprocess.PIPE, check=True)
        except FileNotFoundError:
            raise RuntimeError('Pandoc path "{0}" does not exist'.format(pandoc_path))
        pandoc_version_match = re.search(rb'\d+\.\d+', proc.stdout)
        if not pandoc_version_match:
            raise RuntimeError('Could not determine Pandoc version from "{0} --version"; faulty Pandoc installation?'.format(pandoc_path))
        pandoc_version_major, pandoc_version_minor = (float(x) for x in pandoc_version_match.group().split(b'.', 1))
        if pandoc_version_major < 2 or pandoc_version_minor < 4:
            raise RuntimeError('Pandoc at "{0}" is version {1}.{2}, but >= 2.4 is required'.format(pandoc_path, pandoc_version_major, pandoc_version_minor))
        self.pandoc_path = pandoc_path
        if platform.system() == 'Windows':
            pandoc_template_path = pathlib.Path('~/AppData/Roaming/pandoc/templates').expanduser()
        else:
            pandoc_data_path = pathlib.Path(os.environ.get('XDG_DATA_HOME', '~/.local/share')).expanduser() / 'pandoc'
            if not pandoc_data_path.is_dir():
                pandoc_data_path = pathlib.Path('~/.pandoc').expanduser()
            pandoc_template_path = pandoc_data_path / 'templates'
        if not pandoc_template_path.is_dir():
            pandoc_template_path.mkdir(parents=True)
        codebraid_markdown_roundtrip_template_path = pandoc_template_path / 'codebraid_pandoc_markdown_roundtrip.md'
        if not codebraid_markdown_roundtrip_template_path.is_file():
            codebraid_markdown_roundtrip_template_path.write_text(textwrap.dedent('''\
                $if(titleblock)$
                $titleblock$

                $endif$
                $body$
                '''), encoding='utf8')
        self.codebraid_markdown_roundtrip_template_path = codebraid_markdown_roundtrip_template_path

        if not isinstance(pandoc_file_scope, bool):
            raise TypeError
        if pandoc_file_scope and kwargs.get('cross_origin_sessions') is None:
            # `pandoc_file_scope` automatically disables `cross_origin_sessions`
            # unless `cross_origin_sessions` is explicitly specified.
            self.cross_origin_sessions = False
        self.pandoc_file_scope = pandoc_file_scope
        self.other_pandoc_args_at_load = other_pandoc_args_at_load

        if self.from_format == 'markdown' and not pandoc_file_scope and len(self.origins) > 1:
            # If multiple files are being passed to Pandoc for concatenated
            # processing, ensure sufficient whitespace to prevent elements in
            # different files from merging, and insert a comment to prevent
            # indented elements from merging.  This means that the original
            # sources cannot be passed to Pandoc directly.
            for origin_name, origin_string in self.origins.items():
                if origin_string[-1:] == '\n':
                    origin_string += '\n<!--codebraid.eof-->\n\n'
                else:
                    origin_string += '\n\n<!--codebraid.eof-->\n\n'
                self.origins[origin_name] = origin_string
            self.concat_origin_string = ''.join(self.origins.values())

        self.from_format_pandoc_extensions = from_format_pandoc_extensions

        if not isinstance(scroll_sync, bool):
            raise TypeError
        self.scroll_sync = scroll_sync
        if scroll_sync:
            self._io_map = True
            raise NotImplementedError

        self._asts = {}
        self._para_plain_origin_name_node_line_number = []
        self._final_ast = None

        self.codebraid_defaults.update_from_yaml_metadata(next(iter(self.origins.values())))


    from_formats = set(['markdown', 'commonmark_x'])
    multi_origin_formats = set(['markdown', 'commonmark_x'])
    to_formats = None


    def _split_format_extensions(self, format_extensions):
        '''
        Split a Pandoc `--from` or `--to` string of the form
        `format+extension1-extension2` into the format and extensions.
        '''
        if format_extensions is None:
            return (None, None)
        index = min([x for x in (format_extensions.find('+'), format_extensions.find('-')) if x >= 0], default=-1)
        if index >= 0:
            return (format_extensions[:index], format_extensions[index:])
        return (format_extensions, None)


    # Node sets are based on pandocfilters
    # https://github.com/jgm/pandocfilters/blob/master/pandocfilters.py
    _null_block_node_type = set([''])
    _block_node_types = set(['Plain', 'Para', 'CodeBlock', 'RawBlock',
                             'BlockQuote', 'OrderedList', 'BulletList',
                             'DefinitionList', 'Header', 'HorizontalRule',
                             'Table', 'Div']) | _null_block_node_type


    def _run_pandoc(self, *,
                    from_format: str,
                    to_format: Optional[str],
                    from_format_pandoc_extensions: Optional[str]=None,
                    to_format_pandoc_extensions: Optional[str]=None,
                    file_scope=False,
                    input: Optional[Union[str, bytes]]=None,
                    input_paths: Optional[Union[pathlib.Path, Sequence[pathlib.Path]]]=None,
                    input_name: Optional[str]=None,
                    output_path: Optional[pathlib.Path]=None,
                    standalone: bool=False,
                    newline_lf: bool=False,
                    preserve_tabs: bool=False,
                    other_pandoc_args: Optional[List[str]]=None):
        '''
        Convert between formats using Pandoc.

        Communication with Pandoc is accomplished via pipes.
        '''
        if from_format_pandoc_extensions is None:
            from_format_pandoc_extensions = ''
        if to_format_pandoc_extensions is None:
            to_format_pandoc_extensions = ''
        if input and input_paths:
            raise TypeError
        cmd_list = [str(self.pandoc_path),
                    '--from', from_format + from_format_pandoc_extensions]
        if newline_lf:
            cmd_list.extend(['--eol', 'lf'])
        if to_format is not None:
            cmd_list.extend(['--to', to_format + to_format_pandoc_extensions])
        if standalone:
            cmd_list.append('--standalone')
            if to_format is None:
                if output_path is not None and output_path.suffix in ('.md', '.markdown'):
                    cmd_list.extend(['--template', self.codebraid_markdown_roundtrip_template_path.as_posix()])
            elif to_format == 'markdown':
                cmd_list.extend(['--template', self.codebraid_markdown_roundtrip_template_path.as_posix()])
        if file_scope:
            cmd_list.append('--file-scope')
        if preserve_tabs:
            cmd_list.append('--preserve-tabs')
        if output_path:
            cmd_list.extend(['--output', output_path.as_posix()])
        if other_pandoc_args:
            if not newline_lf:
                cmd_list.extend(other_pandoc_args)
            else:
                eol = False
                for arg in other_pandoc_args:
                    if arg == '--eol':
                        eol = True
                        continue
                    if eol:
                        eol = False
                        continue
                    cmd_list.append(arg)
        if input_paths is not None:
            if isinstance(input_paths, pathlib.Path):
                cmd_list.append(input_paths.as_posix())
            else:
                cmd_list.extend([p.as_posix() for p in input_paths])

        if platform.system() == 'Windows':
            # Prevent console from appearing for an instant
            startupinfo = subprocess.STARTUPINFO()
            startupinfo.dwFlags |= subprocess.STARTF_USESHOWWINDOW
        else:
            startupinfo = None

        if isinstance(input, str):
            input = input.encode('utf8')

        try:
            proc = subprocess.run(cmd_list,
                                  input=input,
                                  stdout=subprocess.PIPE,
                                  stderr=subprocess.PIPE,
                                  startupinfo=startupinfo, check=True)
        except subprocess.CalledProcessError as e:
            if input_paths is not None and input_name is None:
                if isinstance(input_paths, pathlib.Path):
                    input_name = "{0}".format(input_paths.as_posix())
                else:
                    input_name = ', '.join("{0}".format(p.as_posix()) for p in input_paths)
            if input_name is None:
                message = 'Failed to run Pandoc:\n{0}'.format(e.stderr.decode('utf8'))
            else:
                message = 'Failed to run Pandoc on source(s) {0}:\n{1}'.format(input_name, e.stderr.decode('utf8'))
            raise PandocError(message)
        return (proc.stdout, proc.stderr)


    _walk_node_list = staticmethod(walk_node_list)

    def _walk_ast(self, ast, type_filter=None, skip_note_contents=False):
        '''
        Walk all nodes in AST.
        '''
        ast_root_node_list = ast['blocks']
        yield from self._walk_node_list(ast_root_node_list, ast, type_filter=type_filter, skip_note_contents=skip_note_contents)


    @staticmethod
    def _io_map_span_node(origin_name, line_number):
        '''
        Create an empty span node containing source name and line number as
        attributes.  This is used to attach source info to an AST location,
        and then track that location through AST transformations and
        conversions.
        '''
        span_node = {'t': 'Span',
                     'c': [
                              [
                                  '',  # id
                                  ['codebraid--temp'],  # classes
                                  [['trace', '{0}:{1}'.format(origin_name, line_number)]]  # kv pairs
                              ],
                              []  # contents
                          ]
                    }
        return span_node

    @staticmethod
    def _io_map_span_node_to_raw_tracker(span_node):
        span_node['t'] = 'RawInline'
        span_node['c'] = [None, '\x02CodebraidTrace({0})\x03'.format(span_node['c'][0][2][0][1])]


    @staticmethod
    def _freeze_raw_node(node, origin_name, line_number,
                         type_translation_dict={'RawBlock': 'CodeBlock', 'RawInline': 'Code'}):
        '''
        Convert a raw node into a special code node.  This prevents the raw
        node from being prematurely interpreted/discarded during intermediate
        AST transformations.
        '''
        node['t'] = type_translation_dict[node['t']]
        raw_format, raw_content = node['c']
        node['c'] = [
                        [
                            '',  # id
                            ['codebraid--temp'],  # classes
                            [['format', raw_format]],  # kv pairs
                        ],
                        raw_content
                    ]

    @staticmethod
    def _freeze_raw_node_io_map(node, origin_name, line_number,
                                type_translation_dict={'RawBlock': 'CodeBlock', 'RawInline': 'Code'}):
        '''
        Same as `_freeze_raw_node()`, but also store trace info.
        '''
        node['t'] = type_translation_dict[node['t']]
        raw_format, raw_content = node['c']
        node['c'] = [
                        [
                            '',  # id
                            ['codebraid--temp'],  # classes
                            [['format', raw_format], ['trace', '{0}:{1}'.format(origin_name, line_number)]]  # kv pairs
                        ],
                        raw_content
                    ]

    @staticmethod
    def _thaw_raw_node(node,
                       type_translation_dict={'CodeBlock': 'RawBlock', 'Code': 'RawInline'}):
        '''
        Convert a special code node back into its original form as a raw node.
        '''
        node['t'] = type_translation_dict[node['t']]
        node['c'] = [node['c'][0][2][0][1], node['c'][1]]

    @staticmethod
    def _thaw_raw_node_io_map(node,
                              type_translation_dict={'CodeBlock': 'RawBlock', 'Code': 'RawInline'}):
        '''
        Same as `_thaw_raw_node()`, but also return trace info.
        '''
        node['t'] = type_translation_dict[node['t']]
        for k, v in node['c'][0][2]:
            if k == 'format':
                raw_format = v
            elif k == 'trace':
                trace = v
            else:
                raise ValueError
        node['c'] = [raw_format, '\x02CodebraidTrace({0})\x03'.format(trace) + node['c'][1]]


    @staticmethod
    def _find_cb_command(
            text: str,
            *,
            from_format: str,
            max_command_len=max(len(x) for x in CodeChunk.commands),
            commands_set=CodeChunk.commands,
            before_attr_chars=set('{ \t\n'),
            after_attr_chars=set('} \t\n')
        ) -> List[Tuple[int, int, bool, Optional[str]]]:
        '''
        Find all occurrences of `.cb.` and/or `.cb-` in a string, which may
        indicate the presence of Codebraid code.  Try to extract a command
        `.cb<sep><command>` for each occurrence, and if a valid command is not
        found use None.  Check the characters immediately before/after
        `.cb<sep><command>` to determine whether this may be part of code
        attributes.

        Return a list of tuples.  Each tuple contains data about an occurrence
        of `.cb<sep>`:
            (<index>, <line_number>, <maybe_codebraid_attr>, <command>)
        '''
        command_prefixes = {'.cb-': True}
        if not from_format.startswith('commonmark_x'):
            command_prefixes['.cb.'] = True
        results = []
        line_num = 1
        start_index = 0
        while True:
            maybe_start_cb_command_index = -1
            for command_prefix, prefix_is_active in command_prefixes.items():
                if prefix_is_active:
                    search_index = text.find(command_prefix, start_index)
                    if search_index == -1:
                        command_prefixes[command_prefix] = False
                    elif maybe_start_cb_command_index == -1 or search_index < maybe_start_cb_command_index:
                        maybe_start_cb_command_index = search_index
            if maybe_start_cb_command_index == -1:
                break
            line_num += text.count('\n', start_index, maybe_start_cb_command_index)
            maybe_end_cb_command_index = maybe_start_cb_command_index + 4
            for _ in range(max_command_len + 1):
                if text[maybe_end_cb_command_index:maybe_end_cb_command_index+1] in after_attr_chars:
                    break
                maybe_end_cb_command_index += 1
            else:
                maybe_end_cb_command_index = -1
            if maybe_end_cb_command_index == -1:
                command = None
            else:
                command = text[maybe_start_cb_command_index+4:maybe_end_cb_command_index]
                if command not in commands_set:
                    command = None
            if command is None:
                start_index = maybe_start_cb_command_index + 4
            else:
                start_index = maybe_end_cb_command_index
            if maybe_start_cb_command_index < 4 or text[maybe_start_cb_command_index-1] not in before_attr_chars:
                # For Codebraid attr, `.cb.<command>` must always have at
                # least 4 characters in front of it, either something like
                # "`x`{.cb.command}" for inline or "```{.cb.command}" for
                # block.  There must always be a "{" or whitespace in front of
                # `.cb.<command>`.
                maybe_codebraid_attr = False
            elif command is not None and text[maybe_end_cb_command_index:maybe_end_cb_command_index+1] not in after_attr_chars:
                # For Codebraid attr with a valid Codebraid command,
                # `.cb.<command>` must be followed by whitespace or a "}".
                maybe_codebraid_attr = False
            else:
                # It would be possible to introduce additional checks here to
                # handle the case of invalid commands that match Pandoc attr
                # syntax, but that would probably introduce significant
                # complexity for marginal benefit.
                maybe_codebraid_attr = True
            results.append((maybe_start_cb_command_index, line_num, maybe_codebraid_attr, command))
        return results


    def _load_and_process_initial_ast(self, *,
                                      origin_string, single_origin_name=None):
        '''
        Convert source string into a Pandoc AST and perform a number of
        operations on the AST.
          * Assign source line numbers to some nodes.  This is needed later
            for syncing error messages and is also needed for SyncTeX when the
            output is LaTeX.
          * Locate all code nodes for further processing.
          * Convert all raw nodes into specially marked code nodes.  This
            allows them to pass through later Pandoc conversions without
            being lost or being interpreted before the final conversion.
            These special code nodes are converted back into raw nodes in the
            final AST before the final format conversion.
        '''
        # Convert source string to AST with Pandoc.
        # Order of extensions is important: earlier override later.
        from_format_pandoc_extensions = ['-smart']
        if self.from_format == 'markdown':
            from_format_pandoc_extensions.append('-latex_macros')
        from_format_pandoc_extensions = ''.join(from_format_pandoc_extensions)

        if self.from_format_pandoc_extensions is not None:
            from_format_pandoc_extensions += self.from_format_pandoc_extensions

        stdout_bytes, stderr_bytes = self._run_pandoc(input=origin_string,
                                                      input_name=single_origin_name,
                                                      from_format=self.from_format,
                                                      from_format_pandoc_extensions=from_format_pandoc_extensions,
                                                      other_pandoc_args=self.other_pandoc_args_at_load,
                                                      to_format='json',
                                                      newline_lf=True,
                                                      preserve_tabs=True)
        if self.from_format_pandoc_extensions is not None:
            self.from_format_pandoc_extensions = self.from_format_pandoc_extensions.replace('+sourcepos', '')

        if stderr_bytes:
            sys.stderr.buffer.write(stderr_bytes)
        try:
            ast = json.loads(stdout_bytes)
        except Exception as e:
            raise PandocError('Failed to load AST (incompatible Pandoc version?):\n{0}'.format(e))
        if not (isinstance(ast, dict) and
                'pandoc-api-version' in ast and isinstance(ast['pandoc-api-version'], list) and
                all(isinstance(x, int) for x in ast['pandoc-api-version']) and 'blocks' in ast):
            raise PandocError('Unrecognized AST format (incompatible Pandoc version?)')
        for k in ('codebraid', 'codebraid_'):
            try:
                del ast['meta'][k]
            except KeyError:
                pass
        self._asts[single_origin_name] = ast


        # Locate all code nodes to find Codebraid code nodes.  Also locate all
        # raw nodes so that they can be "frozen" to avoid changes during
        # intermediate AST manipulations.
        code_raw_node_tuples = list(self._walk_ast(ast, type_filter=pandoc_code_raw_nodes))
        # Create code chunks.  Determine if any are in notes, in which case
        # the order of appearance in the source may be different from that in
        # the AST.
        code_chunks_in_notes = False
        codebraid_node_set = set()
        for node_tuple in code_raw_node_tuples:
            node, parent_node, parent_node_list, parent_node_list_index, in_note = node_tuple
            node_type = node['t']
            if node_type.startswith('Code'):
                if any(c.startswith('cb.') or c.startswith('cb-')for c in node['c'][0][1]):
                    if in_note:
                        code_chunks_in_notes = True
                    code_chunk = PandocCodeChunk(node, parent_node, parent_node_list, parent_node_list_index)
                    self.code_chunks.append(code_chunk)
                    codebraid_node_set.add(id(node))
        # Locate all occurrences of `.cb.` in source(s), to provide
        # traceback information for source errors
        if single_origin_name is not None:
            sources_cb = [(single_origin_name, cb_i) for cb_i in self._find_cb_command(origin_string, from_format=self.from_format)]
            origin_names_stack = [single_origin_name]
        else:
            sources_cb = [(src_name, cb_i) for (src_name, src_string) in self.origins.items() for cb_i in self._find_cb_command(src_string, from_format=self.from_format)]
            origin_names_stack = [k for k in self.origins]
        sources_cb.reverse()
        origin_names_stack.reverse()
        # Process code and raw nodes:  Attach location info to code nodes and
        # "freeze" raw nodes.  If necessary, also search raw node contents to
        # eliminate false positives for code node location.
        if self._io_map:
            raise NotImplementedError
        else:
            freeze_raw_node = self._freeze_raw_node
        if len(sources_cb) == len(self.code_chunks):
            # All `.cb.` in source(s) are active Codebraid code chunks, but
            # may be out of order due to notes.
            skipped = []
            for chunk in self.code_chunks:
                src_cb_i = sources_cb.pop()
                src_name, (_, src_line_number, _, src_command) = src_cb_i
                if chunk.command != src_command:
                    while chunk.command != src_command:
                        skipped.append(src_cb_i)
                        src_cb_i = sources_cb.pop()
                        src_name, (_, src_line_number, _, src_command) = src_cb_i
                chunk.origin_name = src_name
                if chunk.inline:
                    chunk.origin_start_line_number = src_line_number
                else:
                    chunk.origin_start_line_number = src_line_number + 1
                if skipped:
                    while skipped:
                        sources_cb.append(skipped.pop())
            for node_tuple in code_raw_node_tuples:
                node, parent_node, parent_node_list, parent_node_list_index, in_note = node_tuple
                node_type = node['t']
                if node_type.startswith('Raw'):
                    if single_origin_name is None:
                        node_format, node_contents = node['c']
                        node_format = node_format.lower()
                        if node_format == 'html' and node_contents == '<!--codebraid.eof-->':
                            node['t'] = 'Null'
                            del node['c']
                            continue
                    freeze_raw_node(node, '', 0)
        else:
            # Some `.cb.` in source(s) are not active Codebraid code chunks.
            # They are false positives or commented-out Codebraid code chunks.
            # If notes are present, nodes may be out of order.
            code_chunk_iter = iter(self.code_chunks)
            skipped = []
            for node_tuple in code_raw_node_tuples:
                node, parent_node, parent_node_list, parent_node_list_index, in_note = node_tuple
                node_type = node['t']
                if single_origin_name is None and node_type.startswith('Raw'):
                    node_format, node_contents = node['c']
                    node_format = node_format.lower()
                    if node_format == 'html' and node_contents == '<!--codebraid.eof-->':
                        node['t'] = 'Null'
                        del node['c']
                        current_origin_name = origin_names_stack.pop()
                        while sources_cb and sources_cb[-1][0] == current_origin_name:
                            sources_cb.pop()
                        continue
                node_cb = self._find_cb_command(node['c'][1], from_format=self.from_format)
                if node_cb and not node_type.endswith('Block'):
                    # Inline:  code comes before attr
                    for _, _, node_maybe_codebraid, node_command in node_cb:
                        src_cb_i = sources_cb.pop()
                        _, (_, _, _, src_command) = src_cb_i
                        # Can't compare `maybe_codebraid` because Pandoc
                        # strips leading/trailing whitespace from inline code
                        if node_command != src_command:
                            while node_command != src_command:
                                skipped.append(src_cb_i)
                                src_cb_i = sources_cb.pop()
                                _, (_, _, _, src_command) = src_cb_i
                    if skipped:
                        while skipped:
                            sources_cb.append(skipped.pop())
                if id(node) in codebraid_node_set:
                    chunk = next(code_chunk_iter)
                    src_cb_i = sources_cb.pop()
                    src_name, (_, src_line_number, src_maybe_codebraid, src_command) = src_cb_i
                    if chunk.command != src_command or not src_maybe_codebraid:
                        while chunk.command != src_command or not src_maybe_codebraid:
                            skipped.append(src_cb_i)
                            src_cb_i = sources_cb.pop()
                            src_name, (_, src_line_number, src_maybe_codebraid, src_command) = src_cb_i
                    chunk.origin_name = src_name
                    if chunk.inline:
                        chunk.origin_start_line_number = src_line_number
                    else:
                        chunk.origin_start_line_number = src_line_number + 1
                    if skipped:
                        while skipped:
                            sources_cb.append(skipped.pop())
                if node_cb and node_type.endswith('Block'):
                    # Block:  code comes after attr
                    for _, _, node_maybe_codebraid, node_command in node_cb:
                        src_cb_i = sources_cb.pop()
                        _, (_, _, src_maybe_codebraid, src_command) = src_cb_i
                        if node_command != src_command or node_maybe_codebraid != src_maybe_codebraid:
                            while node_command != src_command or node_maybe_codebraid != src_maybe_codebraid:
                                skipped.append(src_cb_i)
                                src_cb_i = sources_cb.pop()
                                _, (_, _, src_maybe_codebraid, src_command) = src_cb_i
                    if skipped:
                        while skipped:
                            sources_cb.append(skipped.pop())
                if node_type.startswith('Raw'):
                    freeze_raw_node(node, '', 0)


    def _extract_code_chunks(self):
        if self.pandoc_file_scope or len(self.origins) == 1:
            for origin_name, origin_string in self.origins.items():
                self._load_and_process_initial_ast(origin_string=origin_string, single_origin_name=origin_name)
        else:
            self._load_and_process_initial_ast(origin_string=self.concat_origin_string)


    def _generate_final_ast(self):
        for code_chunk in reversed(self.code_chunks):
            # Substitute code output into AST in reversed order to preserve
            # indices
            code_chunk.update_parent_node()
        if self._io_map:
            # Insert tracking spans if needed
            io_map_span_node = self._io_map_span_node
            for origin_name, node, line_number in self._para_plain_origin_name_node_line_number:
                if node['t'] in ('Para', 'Plain'):
                    # When `example` is in use, a node can be converted into
                    # a different type
                    node['c'].insert(0, io_map_span_node(origin_name, line_number))

        # Convert modified AST to markdown, then back, so that raw output
        # can be reinterpreted as markdown.
        # Order of extensions is important: earlier override later.
        processed_to_format_extensions = ['-raw_attribute', '-smart']
        if self.from_format == 'markdown':
            processed_to_format_extensions.append('-latex_macros')
        processed_to_format_extensions = ''.join(processed_to_format_extensions)

        processed_markup = collections.OrderedDict()
        for origin_name, ast in self._asts.items():
            markup_bytes, stderr_bytes = self._run_pandoc(input=json.dumps(ast),
                                                          from_format='json',
                                                          to_format='markdown',
                                                          to_format_pandoc_extensions=processed_to_format_extensions,
                                                          standalone=True,
                                                          newline_lf=True,
                                                          preserve_tabs=True)
            if stderr_bytes:
                sys.stderr.buffer.write(stderr_bytes)
            processed_markup[origin_name] = markup_bytes

        if not self.pandoc_file_scope or len(self.origins) == 1:
            for markup_bytes in processed_markup.values():
                final_ast_bytes, stderr_bytes = self._run_pandoc(input=markup_bytes,
                                                                 from_format='markdown',
                                                                 from_format_pandoc_extensions=self.from_format_pandoc_extensions,
                                                                 to_format='json',
                                                                 newline_lf=True,
                                                                 preserve_tabs=True)
                if stderr_bytes:
                    sys.stderr.buffer.write(stderr_bytes)
        else:
            with tempfile.TemporaryDirectory() as tempdir:
                tempdir_path = pathlib.Path(tempdir)
                tempfile_paths = []
                random_suffix = util.random_ascii_lower_alpha(16)
                for n, markup_bytes in enumerate(processed_markup.values()):
                    tempfile_path = tempdir_path / 'codebraid_intermediate_{0}_{1}.txt'.format(n, random_suffix)
                    tempfile_path.write_bytes(markup_bytes)
                    tempfile_paths.append(tempfile_path)
                final_ast_bytes, stderr_bytes = self._run_pandoc(input_paths=tempfile_paths,
                                                                 from_format='markdown',
                                                                 from_format_pandoc_extensions=self.from_format_pandoc_extensions,
                                                                 to_format='json',
                                                                 newline_lf=True,
                                                                 preserve_tabs=True)
                if stderr_bytes:
                    sys.stderr.buffer.write(stderr_bytes)
        final_ast = json.loads(final_ast_bytes)
        self._final_ast = final_ast

        if not self._io_map:
            thaw_raw_node = self._thaw_raw_node
            for node_tuple in self._walk_ast(final_ast):
                node, parent_node, parent_node_list, parent_node_list_index, in_note = node_tuple
                node_type = node['t']
                if node_type in ('Code', 'CodeBlock') and 'codebraid--temp' in node['c'][0][1]:
                    thaw_raw_node(node)
        else:
            io_tracker_nodes = []
            io_map_span_node_to_raw_tracker = self._io_map_span_node_to_raw_tracker
            thaw_raw_node = self._thaw_raw_node_io_map
            for node_tuple in self._walk_ast(final_ast):
                node, parent_node_list, parent_node_list_index = node_tuple
                node_type = node['t']
                if node_type == 'Span' and 'codebraid--temp' in node['c'][0][1]:
                    io_map_span_node_to_raw_tracker(node)
                    io_tracker_nodes.append(node)
                if node_type in ('Code', 'CodeBlock') and 'codebraid--temp' in node['c'][0][1]:
                    thaw_raw_node(node)
            self._io_tracker_nodes = io_tracker_nodes


    def _convert(self, *, to_format, output_path=None, overwrite=False,
                standalone=None, other_pandoc_args=None):
        if self._final_ast is None:
            self._generate_final_ast()

        if to_format is None:
            if output_path is None:
                raise ValueError('Explicit output format is required when it cannot be inferred from output file name')
            to_format_pandoc_extensions = None
        else:
            if not isinstance(to_format, str):
                raise TypeError
            to_format, to_format_pandoc_extensions = self._split_format_extensions(to_format)
        if not isinstance(standalone, bool):
            raise TypeError
        if output_path is not None:
            if not isinstance(output_path, pathlib.Path):
                if isinstance(output_path, str):
                    output_path = pathlib.Path(output_path)
                else:
                    raise TypeError

        if not isinstance(overwrite, bool):
            raise TypeError
        if output_path is not None and not overwrite and output_path.is_file():
            raise RuntimeError('Output path "{0}" exists, but overwrite=False'.format(output_path))

        if not self._io_map:
            converted_bytes, stderr_bytes = self._run_pandoc(input=json.dumps(self._final_ast),
                                                             from_format='json',
                                                             to_format=to_format,
                                                             to_format_pandoc_extensions=to_format_pandoc_extensions,
                                                             standalone=standalone,
                                                             output_path=output_path,
                                                             other_pandoc_args=other_pandoc_args,
                                                             preserve_tabs=True)
            if stderr_bytes:
                sys.stderr.buffer.write(stderr_bytes)
            if output_path is None:
                sys.stdout.buffer.write(converted_bytes)
        else:
            for node in self._io_tracker_nodes:
                node['c'][0] = to_format
            converted_bytes, stderr_bytes = self._run_pandoc(input=json.dumps(self._final_ast),
                                                             from_format='json',
                                                             to_format=to_format,
                                                             to_format_pandoc_extensions=to_format_pandoc_extensions,
                                                             standalone=standalone,
                                                             newline_lf=True,
                                                             other_pandoc_args=other_pandoc_args,
                                                             preserve_tabs=True)
            if stderr_bytes:
                sys.stderr.buffer.write(stderr_bytes)
            converted_lines = util.splitlines_lf(converted_bytes.decode(encoding='utf8')) or ['']
            converted_to_source_dict = {}
            trace_re = re.compile(r'\x02CodebraidTrace\(.+?:\d+\)\x03')
            for index, line in enumerate(converted_lines):
                if '\x02' in line:
                    #  Tracking format:  '\x02CodebraidTrace({0})\x03'
                    line_split = line.split('\x02CodebraidTrace(', 1)
                    if len(line_split) == 1:
                        continue
                    line_before, trace_and_line_after = line_split
                    trace, line_after = trace_and_line_after.split(')\x03', 1)
                    line = line_before + line_after
                    converted_to_source_dict[str(index + 1)] = trace
                    if '\x02' in line:
                        line = trace_re.sub('', line)
                    converted_lines[index] = line
            converted_lines[-1] = converted_lines[-1] + '\n'
            converted = '\n'.join(converted_lines)
            if self.synctex:
                self._save_synctex_data(converted_to_source_dict)
            if output_path is not None:
                output_path.write_text(converted, encoding='utf8')
            else:
                sys.stdout.buffer.write(converted.encode('utf8'))



# -*- coding: utf-8 -*-
#
# Copyright (c) 2021-2022, Geoffrey M. Poore
# All rights reserved.
#
# Licensed under the BSD 3-Clause License:
# http://opensource.org/licenses/BSD-3-Clause
#


from __future__ import annotations


import io
import os
import platform
import sys
from typing import Callable, Optional, TextIO


class Terminal(object):
    '''
    Utils for working with terminal to create output messages.
    '''
    def __init__(self, stream: TextIO):
        if not isinstance(stream, io.TextIOWrapper):
            raise TypeError
        if stream not in (sys.stdout, sys.stderr):
            raise ValueError

        self.stream: io.TextIOWrapper = stream

        self.isatty: bool = False
        self.columns: Callable[[], int] = self._columns_default
        self.supports_ansi_escape_codes: bool = False
        self.clearline: Callable[[Optional[int]], str] = self._clearline_plain
        self.fmt_ok: Callable[[str], str] = self._fmt_ok_plain
        self.fmt_ok_heading: Callable[[str], str] = self._fmt_ok_heading_plain
        self.fmt_error: Callable[[str], str] = self._fmt_error_plain
        self.fmt_error_heading: Callable[[str], str] = self._fmt_error_heading_plain
        self.fmt_warning: Callable[[str], str] = self._fmt_warning_plain
        self.fmt_warning_heading: Callable[[str], str] = self._fmt_warning_heading_plain
        self.fmt_notify: Callable[[str], str] = self._fmt_notify_plain
        self.fmt_notify_heading: Callable[[str], str] = self._fmt_notify_heading_plain
        self.fmt_delim: Callable[[str], str] = self._fmt_delim_plain

        if not stream.isatty():
            return

        get_terminal_size = os.get_terminal_size
        fileno = stream.fileno()
        if platform.system() == 'Windows' or platform.system().startswith('CYGWIN_NT'):
            # https://bugs.python.org/issue28654
            try:
                get_terminal_size(fileno)
            except OSError:
                return
            self.isatty = True
            if 'ALACRITTY_LOG' in os.environ or 'WT_SESSION' in os.environ:
                # Alacritty or Windows Terminal (https://github.com/microsoft/terminal/issues/1040)
                self.columns = lambda: get_terminal_size(fileno).columns
                self.supports_ansi_escape_codes = True
            else:
                # Many Windows terminal programs interpret a line consisting
                # of <columns> characters followed by a newline as two lines,
                # one line consisting of <columns> characters and one line
                # that is empty (from the newline).  For some but not all of
                # these, it is possible to use `\b\n` to avoid this.  The
                # simplest workaround is to treat the number of columns as one
                # less than the actual value.
                self.columns = lambda: get_terminal_size(fileno).columns - 1
                if os.environ.get('CONEMUANSI') == 'ON' or os.environ.get('TERM_PROGRAM') == 'mintty':
                    # ConEmu or Cmder; or Cygwin, Git for Windows BASH, MSYS, etc.
                    self.supports_ansi_escape_codes = True
            if self.supports_ansi_escape_codes:
                self.clearline = self._clearline_ansi_term
            else:
                self.clearline = self._clearline_win_term
        else:
            self.isatty = True
            self.columns = lambda: get_terminal_size(fileno).columns
            self.supports_ansi_escape_codes = True
            self.clearline = self._clearline_ansi_term
        if self.supports_ansi_escape_codes:
            self.fmt_ok = self._fmt_ok_ansi_color
            self.fmt_ok_heading = self._fmt_ok_heading_ansi_color
            self.fmt_error = self._fmt_error_ansi_color
            self.fmt_error_heading = self._fmt_error_heading_ansi_color
            self.fmt_warning = self._fmt_warning_ansi_color
            self.fmt_warning_heading = self._fmt_warning_heading_ansi_color
            self.fmt_notify = self._fmt_notify_ansi_color
            self.fmt_notify_heading = self._fmt_notify_heading_ansi_color
            self.fmt_delim = self._fmt_delim_ansi_color


    @staticmethod
    def _columns_default():
        return 80


    def _clearline_plain(self, columns: Optional[int]=None) -> str:
        return '\n'

    def _clearline_ansi_term(self, columns: Optional[int]=None) -> str:
        return '\x1b[2K\r'

    def _clearline_win_term(self, columns: Optional[int]=None) -> str:
        if columns is None:
            columns = self.columns()
        return '\r' + ' '*columns + '\r'


    @staticmethod
    def _fmt_ok_plain(string: str) -> str:
        return string

    @staticmethod
    def _fmt_ok_ansi_color(string: str) -> str:
        return f'\x1b[92m{string}\x1b[0m'

    @staticmethod
    def _fmt_ok_heading_plain(string: str) -> str:
        return string

    @staticmethod
    def _fmt_ok_heading_ansi_color(string: str) -> str:
        return f'\x1b[92;1m{string}\x1b[0m'


    @staticmethod
    def _fmt_error_plain(string: str) -> str:
        return string

    @staticmethod
    def _fmt_error_ansi_color(string: str) -> str:
        return f'\x1b[91m{string}\x1b[0m'

    @staticmethod
    def _fmt_error_heading_plain(string: str) -> str:
        return string

    @staticmethod
    def _fmt_error_heading_ansi_color(string: str) -> str:
        return f'\x1b[91;1m{string}\x1b[0m'


    @staticmethod
    def _fmt_warning_plain(string: str) -> str:
        return string

    @staticmethod
    def _fmt_warning_ansi_color(string: str) -> str:
        return f'\x1b[93m{string}\x1b[0m'

    @staticmethod
    def _fmt_warning_heading_plain(string: str) -> str:
        return string

    @staticmethod
    def _fmt_warning_heading_ansi_color(string: str) -> str:
        return f'\x1b[93;1m{string}\x1b[0m'


    @staticmethod
    def _fmt_notify_plain(string: str) -> str:
        return string

    @staticmethod
    def _fmt_notify_ansi_color(string: str) -> str:
        return f'\x1b[96m{string}\x1b[0m'

    @staticmethod
    def _fmt_notify_heading_plain(string: str) -> str:
        return string

    @staticmethod
    def _fmt_notify_heading_ansi_color(string: str) -> str:
        return f'\x1b[96;1m{string}\x1b[0m'


    @staticmethod
    def _fmt_delim_plain(string: str) -> str:
        return string

    @staticmethod
    def _fmt_delim_ansi_color(string: str) -> str:
        return f'\x1b[94m{string}\x1b[0m'


stdout_term = Terminal(sys.stdout)
stderr_term = Terminal(sys.stderr)



# -*- coding: utf-8 -*-
#
# Copyright (c) 2019-2022, Geoffrey M. Poore
# All rights reserved.
#
# Licensed under the BSD 3-Clause License:
# http://opensource.org/licenses/BSD-3-Clause
#


from __future__ import annotations


import argparse
import io
import json
import pathlib
import sys
from . import converters
from . import err
from . import util
from .codebraid_defaults import CodebraidDefaults
from .version import __version__ as version




def main():
    class ArgumentParser(argparse.ArgumentParser):
        '''
        Custom argument parser that lists "[<PANDOC OPTIONS>]" in help,
        instead of either leaving them off altogether due to
        argparse.SUPPRESS, or listing them all and making help unnecessarily
        verbose.
        '''
        def print_help(self):
            original_stdout = sys.stdout
            try:
                temp_stdout = io.StringIO()
                sys.stdout = temp_stdout
                super().print_help()
                help_text = temp_stdout.getvalue()
            finally:
                sys.stdout = original_stdout
            try:
                before, after = help_text.split('[FILE', 1)
            except ValueError:
                print(help_text)
            else:
                indent = before.rsplit('\n', 1)[1]
                help_text = before + '[<PANDOC OPTIONS>]\n' + indent + '[FILE' + after
                help_text += '  [<PANDOC OPTIONS>]\n'
                help_text += indent + 'See output of "pandoc --help"\n'
                print(help_text)

    parser = ArgumentParser(prog='codebraid', allow_abbrev=False)
    parser.set_defaults(func=lambda x: parser.print_help())
    parser.add_argument('--version', action='version', version='Codebraid {0}'.format(version))
    subparsers = parser.add_subparsers(dest='subparser_name')

    parser_pandoc = subparsers.add_parser('pandoc',
                                          help='Execute code embedded in a document format supported by Pandoc (requires Pandoc >= 2.4)')
    parser_pandoc.set_defaults(func=pandoc)
    parser_pandoc.add_argument('-f', '--from', '-r', '--read', dest='from_format',
                               help='From format (may include Pandoc format extensions: format+ext1-ext2)')
    parser_pandoc.add_argument('-t', '--to', '-w', '--write', dest='to_format',
                               help='To format (may include Pandoc format extensions: format+ext1-ext2)')
    parser_pandoc.add_argument('-o', '--output',
                               help='File for saving output (otherwise it is written to stdout)')
    parser_pandoc.add_argument('--overwrite',
                               help='Overwrite existing files',
                               action='store_true')
    parser_pandoc.add_argument('-s', '--standalone', action='store_true',
                               help=argparse.SUPPRESS)
    parser_pandoc.add_argument('--file-scope', action='store_true', dest='pandoc_file_scope',
                               help=argparse.SUPPRESS)
    parser_pandoc.add_argument('--no-cache', action='store_true',
                               help='Do not cache code output so that code is always executed during each build '
                                    '(a cache directory may still be created for use with temporary files)')
    parser_pandoc.add_argument('--cache-dir',
                               help='Location for caching code output (default is "_codebraid" in document directory)')
    parser_pandoc.add_argument('--live-output', action='store_true',
                               help='Show code output (stdout and stderr) live in the terminal during code execution. '
                                    'For Jupyter kernels, also show  errors and a summary of rich output. '
                                    'Output still appears in the document as normal. '
                                    'Individual sessions can override this by setting live_output=false in the document.')
    parser_pandoc.add_argument('--no-execute', action='store_true',
                               help='Disable code execution.  Only load code output from cache, if it exists.')
    parser_pandoc.add_argument('--only-code-output', metavar='FORMAT',
                               help='Write code output to stdout in the specified format. '
                                    'Cached output is written immediately. '
                                    'Output from execution is written as soon as it becomes available. '
                                    'No document is created. '
                                    'Options --to and --output are only used (if at all) to inform code output formatting.')
    parser_pandoc.add_argument('--stdin-json-header', action='store_true',
                               help='Treat the first line of stdin as a header in JSON format containing data about stdin such as file origin.')
    parser_pandoc.add_argument('files', nargs='*', metavar='FILE',
                               help="Files (multiple files are allowed for formats supported by Pandoc)")
    for opts_or_long_opt, narg in PANDOC_OPTIONS.items():
        if isinstance(opts_or_long_opt, tuple):
            short_opt, long_opt = opts_or_long_opt
        else:
            short_opt, long_opt = (None, opts_or_long_opt)
        if narg == 0:
            if short_opt is None:
                parser_pandoc.add_argument(long_opt, action='store_true', dest=long_opt,
                                           help=argparse.SUPPRESS)
            else:
                parser_pandoc.add_argument(short_opt, long_opt, action='store_true', dest=long_opt,
                                           help=argparse.SUPPRESS)
        elif narg == 1:
            if short_opt is None:
                parser_pandoc.add_argument(long_opt, dest=long_opt, action='append',
                                           help=argparse.SUPPRESS)
            else:
                parser_pandoc.add_argument(short_opt, long_opt, dest=long_opt, action='append',
                                           help=argparse.SUPPRESS)
        elif narg == '?':
            if short_opt is None:
                parser_pandoc.add_argument(long_opt, nargs='?', const=True, default=None, dest=long_opt,
                                           help=argparse.SUPPRESS)
            else:
                parser_pandoc.add_argument(short_opt, long_opt, nargs='?', const=True, default=None, dest=long_opt,
                                           help=argparse.SUPPRESS)
        else:
            raise ValueError

    args = parser.parse_args()
    args.func(args)




def pandoc(args):
    # Stay consistent with Pandoc's requirement of UTF-8
    sys.stdin = io.TextIOWrapper(sys.stdin.buffer, encoding='utf_8_sig')

    other_pandoc_args = []
    if vars(args).get('--defaults') is not None:
        if args.from_format is None:
            sys.exit('Must specify input format ("--from" or "--read") when using default options ("--defaults")')
        if args.to_format is None:
            sys.exit('Must specify output format ("--to" or "--write") when using default options ("--defaults")')
        if args.output is None:
            sys.exit('Must specify output file ("--output") when using default options ("--defaults"); for stdout, use "-o -"')
    if args.output == '-':
        args.output = None
    for k, v in vars(args).items():
        if isinstance(k, str) and k.startswith('--') and v not in (None, False):
            if isinstance(v, list):
                for v_i in v:
                    other_pandoc_args.append(k)
                    other_pandoc_args.append(v_i)
            elif k in ('--katex', '--mathjax', '--webtex') and not isinstance(v, bool):
                other_pandoc_args.append(f'{k}={v}')
            else:
                other_pandoc_args.append(k)
                if not isinstance(v, bool):
                    other_pandoc_args.append(v)

    if not args.files or (len(args.files) == 1 and args.files[0] == '-'):
        paths = None
        string_origins = None
        try:
            strings = sys.stdin.read()
        except UnicodeDecodeError as e:
            sys.exit(f'Input must be UTF-8:\n{e}')
        if args.stdin_json_header:
            stdin_lines = util.splitlines_lf(strings)
            if not stdin_lines:
                sys.exit('Missing data for --stdin-json-header')
            if len(stdin_lines) == 1:
                stdin_lines.append('')
            json_header_str = stdin_lines[0]
            try:
                json_header = json.loads(json_header_str)
            except Exception as e:
                sys.exit(f'Invalid data for --stdin-json-header:\n{e}')
            if not isinstance(json_header, dict):
                sys.exit(f'Invalid data for --stdin-json-header: expected dict')
            json_header_origins = json_header.get('origins')
            if json_header_origins is not None:
                if not isinstance(json_header_origins, list):
                    sys.exit('Invalid data for --stdin-json-header field "string_origins"')
                start_line = 1
                strings = []
                string_origins = []
                for entry in json_header_origins:
                    origin_name = entry['path']
                    origin_length = entry['lines']
                    string_origins.append(origin_name)
                    strings.append('\n'.join(stdin_lines[start_line:start_line+origin_length]))
                    start_line += origin_length
            else:
                strings = '\n'.join(stdin_lines[1:])
    else:
        paths = args.files
        strings = None
        string_origins = None

    codebraid_defaults = CodebraidDefaults()
    if args.live_output:
        codebraid_defaults['live_output'] = args.live_output

    preview = False
    other_pandoc_args_at_load = None
    for n, arg in enumerate(other_pandoc_args):
        if n > 0 and 'pandoc-sourcepos-sync' in arg and other_pandoc_args[n-1] in ('-L', '--lua-filter'):
            other_pandoc_args_at_load = other_pandoc_args[n-1:n+1]
            other_pandoc_args = other_pandoc_args[:n-1] + other_pandoc_args[n+1:]
            preview = True
            break

    if args.output in (None, '-'):
        output_path = None
    else:
        output_path = pathlib.Path(args.output).expanduser()
        if not args.overwrite and output_path.is_file():
            # There is also a check for this in `converter.convert()`, but should
            # fail early.
            sys.exit(f'File "{args.output}" already exists (to replace it, add option "--overwrite")')
    try:
        with converters.PandocConverter(
            paths=paths,
            strings=strings,
            string_origins=string_origins,
            from_format=args.from_format,
            pandoc_file_scope=args.pandoc_file_scope,
            no_cache=args.no_cache,
            cache_path=args.cache_dir,
            codebraid_defaults=codebraid_defaults,
            other_pandoc_args_at_load=other_pandoc_args_at_load,
            no_execute=args.no_execute,
            only_code_output=args.only_code_output,
        ) as converter:
            if not args.only_code_output:
                converter.convert(
                    to_format=args.to_format,
                    standalone=args.standalone,
                    output_path=output_path,
                    overwrite=args.overwrite,
                    other_pandoc_args=other_pandoc_args
                )
            exit_code = converter.exit_code
    except err.YAMLMetadataError as e:
        sys.exit(str(e))

    sys.exit(exit_code)


PANDOC_OPTIONS  = {
    '--data-dir': 1,
    '--base-header-level': 1,
    '--strip-empty-paragraphs': 0,
    '--indented-code-classes': 1,
    ('-F', '--filter'): 1,
    ('-L', '--lua-filter'): 1,
    '--shift-heading-level-by': 1,
    ('-p', '--preserve-tabs'): 0,
    '--tab-stop': 1,
    '--track-changes': 1,
    '--extract-media': 1,
    '--template': 1,
    ('-M', '--metadata'): 1,
    '--metadata-file': 1,
    ('-d', '--defaults'): 1,
    ('-V', '--variable'): 1,
    '--dpi': 1,
    '--eol': 1,
    '--wrap': 1,
    '--columns': 1,
    '--strip-comments': 0,
    '--toc': 0,
    '--table-of-contents': 0,
    '--toc-depth': 1,
    '--no-highlight': 0,
    '--highlight-style': 1,
    '--syntax-definition': 1,
    ('-H', '--include-in-header'): 1,
    ('-B', '--include-before-body'): 1,
    ('-A', '--include-after-body'): 1,
    '--resource-path': 1,
    '--request-header': 1,
    '--abbreviations': 1,
    '--self-contained': 0,
    '--html-q-tags': 0,
    '--ascii': 0,
    '--reference-links': 0,
    '--reference-location': 1,
    '--atx-headers': 0,
    '--top-level-division': 1,
    ('-N', '--number-sections'): 0,
    '--number-offset': 1,
    '--listings': 0,
    ('-i', '--incremental'): 0,
    '--slide-level': 1,
    '--section-divs': 0,
    '--default-image-extension': 1,
    '--email-obfuscation': 1,
    '--id-prefix': 1,
    ('-T', '--title-prefix'): 1,
    ('-c', '--css'): 1,
    '--reference-doc': 1,
    '--epub-subdirectory': 1,
    '--epub-cover-image': 1,
    '--epub-metadata': 1,
    '--epub-embed-font': 1,
    '--epub-chapter-level': 1,
    '--pdf-engine': 1,
    '--pdf-engine-opt': 1,
    '--ipynb-output': 1,
    ('-C', '--citeproc'): 0,
    '--bibliography': 1,
    '--csl': 1,
    '--citation-abbreviations': 1,
    '--natbib': 0,
    '--biblatex': 0,
    '--mathml': 0,
    '--webtex': '?',
    '--mathjax': '?',
    '--katex': '?',
    '--gladtex': 0,
    '--trace': 0,
    '--dump-args': 0,
    '--ignore-args': 0,
    '--verbose': 0,
    '--quiet': 0,
    '--fail-if-warnings': 0,
    '--log': 1,
    '--bash-completion': 0,
}



# -*- coding: utf-8 -*-
#
# Copyright (c) 2022, Geoffrey M. Poore
# All rights reserved.
#
# Licensed under the BSD 3-Clause License:
# http://opensource.org/licenses/BSD-3-Clause
#


from __future__ import annotations


import re
from typing import Callable
YAML = None
yaml = None
try:
    from ruamel.yaml import YAML
except ImportError:
    try:
        from ruamel_yaml import YAML
    except ImportError:
        pass
if YAML is not None:
    yaml = YAML(typ='safe', pure=True)
from . import err


_yaml_metadata_re = re.compile(r'^---[ \t]*\n(.+?)\n(?:---|\.\.\.)[ \t]*\n', re.DOTALL)


_codebraid_defaults_schema: dict[str, tuple[Callable[[str], bool], str]] = {
    'jupyter': (
        lambda x: (
            isinstance(x, bool) or
            (isinstance(x, dict) and all(k in ('kernel', 'timeout') for k in x) and
                isinstance(x.get('kernel', ''), str) and isinstance(x.get('timeout', 0), int))
        ),
        'bool, or dict containing "kernel" (string) and/or "timeout" (int)'
    ),
    'live_output': (lambda x: isinstance(x, bool), 'bool'),
}
_codebraid_defaults_fallback = {
}
for k, v in _codebraid_defaults_fallback.items():
    if k not in _codebraid_defaults_schema or not _codebraid_defaults_schema[k][0](v):
        raise err.CodebraidError('Invalid defaults fallback')


class CodebraidDefaults(dict):
    def __init__(self):
        super().__init__(_codebraid_defaults_fallback)
        self._key_is_set = set()


    def update(self, data: dict):
        for k, v in data.items():
            self.__setitem__(k, v)


    def __setitem__(self, key, value):
        if key in self._key_is_set:
            return
        try:
            schema_func, value_doc = _codebraid_defaults_schema[key]
        except KeyError:
            raise err.CodebraidError(f'Unknown key "{key}"')
        if not schema_func(value):
            raise err.CodebraidError(f'Incorrect value for key "{key}" (expected {value_doc})')
        super().__setitem__(key, value)
        self._key_is_set.add(key)


    def update_from_yaml_metadata(self, first_origin_string: str):
        if not first_origin_string.startswith('---'):
            return
        match = _yaml_metadata_re.match(first_origin_string)
        if not match or 'codebraid' not in match.group(1):
            return
        if yaml is None:
            raise err.YAMLMetadataError(
                'Cannot load YAML metadata: ruamel.yaml (or ruamel_yaml for conda) is not installed'
            )
        try:
            metadata = yaml.load(match.group(1))
        except Exception as e:
            raise err.YAMLMetadataError(f'Invalid YAML metadata:\n{e}')
        codebraid_keys = ('codebraid', 'codebraid_')
        codebraid_metadata = None
        for key in codebraid_keys:
            data = metadata.get(key)
            if data:
                if codebraid_metadata is not None:
                    raise err.YAMLMetadataError(
                        'Invalid YAML metadata: only a single Codebraid key is permitted '
                        f'''({', '.join(f'"{x}"' for x in codebraid_keys)})'''
                    )
                codebraid_metadata = data
        if codebraid_metadata is None:
            return
        if not isinstance(codebraid_metadata, dict):
            raise err.YAMLMetadataError('Invalid YAML metadata: Codebraid key should map to a dict')
        try:
            self.update(codebraid_metadata)
        except err.CodebraidError as e:
            raise err.YAMLMetadataError(f'Invalid YAML metadata:\n{e}')


    def get_keypath(self, keypath: str, default=None):
        keypath_elems = keypath.split('.')
        obj = self
        for elem in keypath_elems[:-1]:
            try:
                obj = obj[elem]
            except KeyError:
                return default
            if not isinstance(obj, dict):
                return default
        try:
            obj = obj[keypath_elems[-1]]
        except KeyError:
            return default
        return obj



# -*- coding: utf-8 -*-
#
# Copyright (c) 2018-2021, Geoffrey M. Poore
# All rights reserved.
#
# Licensed under the BSD 3-Clause License:
# http://opensource.org/licenses/BSD-3-Clause
#


import sys
if sys.version_info < (3, 7):
    sys.exit('Codebraid requires Python 3.7+')
import pathlib
from setuptools import setup




# Extract the version from version.py, using functions in fmtversion.py
fmtversion_path = pathlib.Path(__file__).parent / 'codebraid' / 'fmtversion.py'
exec(compile(fmtversion_path.read_text(encoding='utf8'), 'codebraid/fmtversion.py', 'exec'))
version_path = pathlib.Path(__file__).parent / 'codebraid' / 'version.py'
version = get_version_from_version_py_str(version_path.read_text(encoding='utf8'))

readme_path = pathlib.Path(__file__).parent / 'README.md'
long_description = readme_path.read_text(encoding='utf8')


setup(name='codebraid',
      version=version,
      py_modules=[],
      packages=[
          'codebraid',
          'codebraid.converters',
          'codebraid.codeprocessors'
      ],
      package_data = {
          'codebraid': ['languages/*.bespon']
      },
      description='Live code in Pandoc Markdown',
      long_description=long_description,
      long_description_content_type='text/markdown',
      author='Geoffrey M. Poore',
      author_email='gpoore@gmail.com',
      url='http://github.com/gpoore/codebraid',
      license='BSD',
      keywords=['dynamic documents', 'reproducible research', 'notebook',
                'markdown', 'pandoc', 'LaTeX'],
      python_requires='>=3.7',
      install_requires=[
          'bespon>=0.6',
      ],
      # https://pypi.python.org/pypi?:action=list_classifiers
      classifiers=[
          'Development Status :: 4 - Beta',
          'Environment :: Console',
          'Intended Audience :: Developers',
          'Intended Audience :: Education',
          'Intended Audience :: Science/Research',
          'License :: OSI Approved :: BSD License',
          'Operating System :: OS Independent',
          'Programming Language :: Python :: 3.7',
          'Programming Language :: Python :: 3.8',
          'Programming Language :: Python :: 3.9',
          'Programming Language :: Python :: 3.10',
          'Topic :: Documentation',
          'Topic :: Education',
          'Topic :: Software Development',
          'Topic :: Software Development :: Build Tools',
          'Topic :: Software Development :: Documentation',
          'Topic :: Text Processing',
          'Topic :: Text Processing :: Markup',
      ],
      entry_points = {
          'console_scripts': ['codebraid = codebraid.cmdline:main'],
      },
)



# -*- coding: utf-8 -*-
#
# Copyright (c) 2021-2022, Geoffrey M. Poore
# All rights reserved.
#
# Licensed under the BSD 3-Clause License:
# http://opensource.org/licenses/BSD-3-Clause
#


from __future__ import annotations


import collections
from collections import OrderedDict as ODict
import pathlib
import re
import textwrap
import typing
from typing import List, NamedTuple, Optional, Union
from . import message
from . import util
if typing.TYPE_CHECKING:
    from .code_collections import CodeCollection




class CodeKey(NamedTuple):
    '''
    Collection of attributes uniquely identifying a session or source.
    '''
    lang: Optional[str]
    name: Optional[str]
    type: str
    origin_name: Optional[str]


class Include(dict):
    '''
    Store code chunk options related to including a file or other external
    resource.  Also perform the include and modify the code chunk as
    necessary.
    '''
    def __init__(self, code_chunk, include_options):
        # Start by creating fallback values for attributes
        self.code_lines = []

        self.code_chunk = code_chunk
        if not isinstance(include_options, dict):
            code_chunk.errors.append(message.SourceError('Invalid "include" value "{0}"'.format(include_options)))
            return

        if not all(k in self.keywords for k in include_options):
            unknown_keys = ', '.join("{0}".format(k) for k in include_options if k not in self.keywords)
            code_chunk.errors.append(message.SourceError('Unknown "include" keywords: {0}'.format(unknown_keys)))
        if not all(isinstance(v, str) and v for v in include_options.values()):
            invalid_value_keys = ', '.join("{0}".format(k) for k, v in include_options.items() if not isinstance(v, str) or not v)
            code_chunk.errors.append(message.SourceError('Invalid values for "include" keywords: {0}'.format(invalid_value_keys)))

        start_keywords = tuple(k for k in include_options if k in self._start_keywords)
        end_keywords = tuple(k for k in include_options if k in self._end_keywords)
        range_keywords = tuple(k for k in include_options if k in self._range_keywords)
        if ((range_keywords and (start_keywords or end_keywords)) or
                len(range_keywords) > 1 or len(start_keywords) > 1 or len(end_keywords) > 1):
            conflicting_keys = ', '.join("{0}".format(k) for k in include_options if k in self._selection_keywords)
            code_chunk.errors.append(message.SourceError('Too many keywords for selecting part of an "include" file: {0}'.format(conflicting_keys)))

        file = include_options.get('file', None)
        encoding = include_options.get('encoding', 'utf8')
        if file is None:
            code_chunk.errors.append(message.SourceError('Missing "include" keyword "file"'))

        if code_chunk.errors:
            return

        file_path = pathlib.Path(file).expanduser()
        try:
            text = file_path.read_text(encoding=encoding)
        except FileNotFoundError:
            code_chunk.errors.append(message.SourceError('Cannot include nonexistent file "{0}"'.format(file)))
        except PermissionError:
            code_chunk.errors.append(message.SourceError('Insufficient permissions to access file "{0}"'.format(file)))
        except LookupError:
            code_chunk.errors.append(message.SourceError('Unknown encoding "{0}"'.format(encoding)))
        except UnicodeDecodeError:
            code_chunk.errors.append(message.SourceError('Cannot decode file "{0}" with encoding "{1}"'.format(file, encoding)))
        if code_chunk.errors:
            return

        selection_keywords = start_keywords + end_keywords + range_keywords
        if selection_keywords:
            for kw in selection_keywords:
                text = getattr(self, '_option_'+kw)(include_options[kw], text)
                if code_chunk.errors:
                    return
        code_lines = util.splitlines_lf(text)
        self.code_lines = code_lines
        self.update(include_options)


    keywords = set(['file', 'encoding', 'lines', 'regex',
                    'start_string', 'start_regex', 'after_string', 'after_regex',
                    'before_string', 'before_regex', 'end_string', 'end_regex'])
    _start_keywords = set(['start_string', 'start_regex', 'after_string', 'after_regex'])
    _end_keywords = set(['before_string', 'before_regex', 'end_string', 'end_regex'])
    _range_keywords = set(['lines', 'regex'])
    _selection_keywords = _start_keywords | _end_keywords | _range_keywords


    def _option_lines(self, value, text,
                      pattern_re=re.compile(r'{n}(?:-(?:{n})?)?(?:,{n}(?:-(?:{n})?)?)*\Z'.format(n='[1-9][0-9]*'))):
        value = value.replace(' ', '')
        if not pattern_re.match(value):
            self.code_chunk.errors.append(message.SourceError('Invalid value for "include" option "lines"'))
            return
        max_line_number = text.count('\n')
        if text[-1:] != '\n':
            max_line_number += 1
        include_line_indices = set()
        for line_range in value.split(','):
            if '-' not in line_range:
                include_line_indices.add(int(line_range)-1)
            else:
                start, end = line_range.split('-')
                start = int(start) - 1
                end = int(end) if end else max_line_number
                include_line_indices.update(range(start, end))
        text_lines = util.splitlines_lf(text)
        return '\n'.join(text_lines[n] for n in sorted(include_line_indices))


    def _option_regex(self, value, text):
        try:
            pattern_re = re.compile(value, re.MULTILINE | re.DOTALL)
        except re.error:
            self.code_chunk.errors.append(message.SourceError('Invalid regex pattern for "include" option "regex"'))
            return
        match = pattern_re.search(text)
        if match is None:
            self.code_chunk.errors.append(message.SourceError('The pattern given by "include" option "regex" was not found'))
            return
        return match.group()


    def _option_start_string(self, value, text):
        index = text.find(value)
        if index < 0:
            self.code_chunk.errors.append(message.SourceError('The pattern given by "include" option "start_string" was not found'))
            return
        return text[index:]


    def _option_start_regex(self, value, text):
        try:
            pattern_re = re.compile(value, re.MULTILINE | re.DOTALL)
        except re.error:
            self.code_chunk.errors.append(message.SourceError('Invalid regex pattern for "include" option "start_regex"'))
            return
        match = pattern_re.search(text)
        if match is None:
            self.code_chunk.errors.append(message.SourceError('The pattern given by "include" option "start_regex" was not found'))
            return
        return text[match.start():]


    def _option_after_string(self, value, text):
        index = text.find(value)
        if index < 0:
            self.code_chunk.errors.append(message.SourceError('The pattern given by "include" option "after_string" was not found'))
            return
        return text[index+len(value):]


    def _option_after_regex(self, value, text):
        try:
            pattern_re = re.compile(value, re.MULTILINE | re.DOTALL)
        except re.error:
            self.code_chunk.errors.append(message.SourceError('Invalid regex pattern for "include" option "after_regex"'))
            return
        match = pattern_re.search(text)
        if match is None:
            self.code_chunk.errors.append(message.SourceError('The pattern given by "include" option "after_regex" was not found'))
            return
        return text[match.end():]


    def _option_before_string(self, value, text):
        index = text.find(value)
        if index < 0:
            self.code_chunk.errors.append(message.SourceError('The pattern given by "include" option "before_string" was not found'))
            return
        return text[:index]


    def _option_before_regex(self, value, text):
        try:
            pattern_re = re.compile(value, re.MULTILINE | re.DOTALL)
        except re.error:
            self.code_chunk.errors.append(message.SourceError('Invalid regex pattern for "include" option "before_regex"'))
            return
        match = pattern_re.search(text)
        if match is None:
            self.code_chunk.errors.append(message.SourceError('The pattern given by "include" option "before_regex" was not found'))
            return
        return text[:match.start()]


    def _option_end_string(self, value, text):
        index = text.find(value)
        if index < 0:
            self.code_chunk.errors.append(message.SourceError('The pattern given by "include" option "end_string" was not found'))
            return
        return text[:index+len(value)]


    def _option_end_regex(self, value, text):
        try:
            pattern_re = re.compile(value, re.MULTILINE | re.DOTALL)
        except re.error:
            self.code_chunk.errors.append(message.SourceError('Invalid regex pattern for "include" option "end_regex"'))
            return
        match = pattern_re.search(text)
        if match is None:
            self.code_chunk.errors.append(message.SourceError('The pattern given by "include" option "end_regex" was not found'))
            return
        return text[:match.end()]




class Options(dict):
    '''
    Store code chunk options.  Also modify the code chunk as necessary based
    on the options.

    Option processing methods check options for validity and process them, but
    do not perform any type conversions.  Any desired type conversions must be
    performed in format-specific subclasses of CodeChunk, which can take into
    account the data types that a given document format allows for options.
    Subclasses must also handle duplicate options, since at this point options
    must have been reduced to a dict.

    The effect of all options is independent of their order.  When two options
    would have an order-dependent effect, only one of them is permitted at a
    time.

    Invalid options related to presentation produce errors that do not prevent
    code execution.  Otherwise, errors prevent execution.  When possible,
    option processing proceeds even after an error, to give a more complete
    error message.
    '''
    def __init__(self, code_chunk, custom_options):
        self.code_chunk = code_chunk
        if code_chunk.inline:
            self.update(self._default_inline_options)
        else:
            self.update(self._default_block_options)
        if code_chunk.execute:
            self['session'] = None
        else:
            self['source'] = None
        self['first_chunk_options'] = {}
        if any(k not in self.keywords for k in custom_options):
            unknown_keys = ', '.join('"{0}"'.format(k) for k in custom_options if k not in self.keywords)
            # Raise an error for unknown options.  There is no way to tell
            # whether an execution or presentation option was intended, so
            # take the safer approach.
            code_chunk.errors.append(message.SourceError('Unknown keywords: {0}'.format(unknown_keys)))
            # Treat received `custom_options` as immutable
            custom_options = {k: v for k, v in custom_options.items() if k in self.keywords}
        self.custom_options = custom_options

        for k, v in custom_options.items():
            if k not in self._after_copy_keywords:
                getattr(self, '_option_'+k)(k, v)

        if not code_chunk.errors and 'copy' not in self:
            # Only handle 'show' and 'hide' if there are no errors so far and
            # there is not a pending 'copy', which for some commands might
            # change `.is_expr` or the defaults for 'show'.  If there are
            # errors, 'show' and 'hide' are never used.
            if code_chunk.inline:
                self['show'] = self._default_inline_show[code_chunk.command].copy()
            else:
                self['show'] = self._default_block_show[code_chunk.command].copy()
            for k, v in custom_options.items():
                if k in self._after_copy_keywords:
                    getattr(self, '_option_'+k)(k, v)


    def finalize_after_copy(self):
        '''
        Complete any option processing that must wait until after copying.
        For the paste command, 'show' can be inherited.  For paste and code,
        `.is_expr` can be inherited.  'lang' can also be inherited.
        '''
        code_chunk = self.code_chunk
        custom_options = self.custom_options
        if self['lang'] is None:
            self['inherited_lang'] = True
            self['lang'] = code_chunk.copy_chunks[0].options['lang']
        if code_chunk.inline:
            if code_chunk.command == 'paste' and 'show' not in custom_options:
                self['show'] = code_chunk.copy_chunks[0].options['show'].copy()  # Inherit
            else:
                self['show'] = self._default_inline_show[code_chunk.command].copy()
        else:
            if code_chunk.command == 'paste' and 'show' not in custom_options:
                self['show'] = code_chunk.copy_chunks[0].options['show'].copy()  # Inherit
            else:
                self['show'] = self._default_block_show[code_chunk.command].copy()

        for key in self._after_copy_keywords:
            if key in custom_options:
                getattr(self, '_option_'+key)(key, custom_options[key])


    _base_keywords = set(['complete', 'copy', 'example', 'hide', 'hide_markup_keys', 'include',
                          'lang', 'name', 'outside_main', 'session', 'source', 'show'])
    _layout_keywords = set(['{0}_{1}'.format(dsp, kw) if dsp else kw
                            for dsp in ('', 'markup', 'copied_markup', 'code', 'stdout', 'stderr')
                            for kw in ('first_number', 'line_numbers', 'rewrap_lines', 'rewrap_width', 'expand_tabs', 'tab_size')])
    _first_chunk_execute_keywords = set(['executable', 'jupyter_kernel'])
    _first_chunk_save_keywords = set(['save', 'save_as'])
    _first_chunk_other_keywords = set(['executable_opts', 'args', 'jupyter_timeout', 'live_output'])
    _first_chunk_keywords = _first_chunk_execute_keywords | _first_chunk_save_keywords | _first_chunk_other_keywords

    keywords = _base_keywords | _layout_keywords | _first_chunk_keywords

    _after_copy_keywords = set(['hide', 'show'])


    # Default values for show and session/source are inserted later based on
    # command and inline status
    _default_inline_options = {'complete': True,
                               'example': False,
                               'lang': None,
                               'inherited_lang': False,
                               'outside_main': False}
    _default_block_options = _default_inline_options.copy()
    _default_block_options.update({'code_first_number': 'next',
                                   'code_line_numbers': True})

    # The defaultdict handles unknown commands that are represented as None
    _default_rich_output = 'latex|markdown|png|jpg|svg|plain'.split('|')
    _default_inline_show = collections.defaultdict(lambda: ODict(),  # Unknown -> show nothing
                                                   {'code':  ODict([('code', 'verbatim')]),
                                                    'expr':  ODict([('expr', 'raw'),
                                                                    ('stderr', 'verbatim')]),
                                                    # expr and rich_output don't clash, because expr is only present
                                                    # with the built-in code execution system, while rich_output
                                                    # requires a Jupyter kernel.  If the built-in system gains
                                                    # rich_output capabilities or there are other related changes,
                                                    # this may need refactoring.
                                                    'nb':    ODict([('expr', 'verbatim'),
                                                                    ('rich_output', _default_rich_output),
                                                                    ('stderr', 'verbatim')]),
                                                    'paste': ODict(),
                                                    'run':   ODict([('stdout', 'raw'),
                                                                    ('stderr', 'verbatim'),
                                                                    ('rich_output', _default_rich_output)])})
    _default_block_show = collections.defaultdict(lambda: ODict(),  # Unknown -> show nothing
                                                  {'code': ODict([('code', 'verbatim')]),
                                                   'nb':   ODict([('code', 'verbatim'),
                                                                  ('stdout', 'verbatim'),
                                                                  ('stderr', 'verbatim'),
                                                                  ('rich_output', _default_rich_output)]),
                                                   'paste': ODict(),
                                                   'repl': ODict([('repl', 'verbatim')]),
                                                   'run':  ODict([('stdout', 'raw'),
                                                                  ('stderr', 'verbatim'),
                                                                  ('rich_output', _default_rich_output)])})


    def _option_bool_can_exec_error(self, key, value):
        if not isinstance(value, bool):
            self.code_chunk.errors.append(message.CanExecSourceError(f'Invalid "{key}" value "{value}"'))
        else:
            self[key] = value

    def _option_bool_error(self, key, value):
        if not isinstance(value, bool):
            self.code_chunk.errors.append(message.SourceError(f'Invalid "{key}" value "{value}"'))
        else:
            self[key] = value

    def _option_str_can_exec_error(self, key, value):
        if not isinstance(value, str):
            self.code_chunk.errors.append(message.CanExecSourceError(f'Invalid "{key}" value "{value}"'))
        else:
            self[key] = value

    def _option_str_error(self, key, value):
        if not isinstance(value, str):
            self.code_chunk.errors.append(message.SourceError(f'Invalid "{key}" value "{value}"'))
        else:
            self[key] = value

    def _option_positive_int_can_exec_error(self, key, value):
        if not isinstance(value, int) or value <= 0:
            self.code_chunk.errors.append(message.CanExecSourceError(f'Invalid "{key}" value "{value}"'))
        else:
            self[key] = value

    def _option_positive_int_error(self, key, value):
        if not isinstance(value, int) or value <= 0:
            self.code_chunk.errors.append(message.SourceError(f'Invalid "{key}" value "{value}"'))
        else:
            self[key] = value

    def _option_first_chunk_bool_error(self, key, value):
        if not isinstance(value, bool):
            self.code_chunk.errors.append(message.SourceError(f'Invalid "{key}" value "{value}"'))
        else:
            self['first_chunk_options'][key] = value

    def _option_first_chunk_string_error(self, key, value):
        if not isinstance(value, str):
            self.code_chunk.errors.append(message.SourceError(f'Invalid "{key}" value "{value}"'))
        else:
            first_chunk_options = self['first_chunk_options']
            if (key in self._first_chunk_execute_keywords and
                    any(k in first_chunk_options for k in self._first_chunk_execute_keywords)):
                conflicting_options = ', '.join('"{0}"'.format(k) for k in self._first_chunk_execute_keywords if k in first_chunk_options)
                self.code_chunk.errors.append(message.SourceError('Conflicting options: {0}'.format(conflicting_options)))
            else:
                first_chunk_options[key] = value

    def _option_first_chunk_int_error(self, key, value):
        if not isinstance(value, int):
            self.code_chunk.errors.append(message.SourceError(f'Invalid "{key}" value "{value}"'))
        else:
            self['first_chunk_options'][key] = value

    _option_executable = _option_first_chunk_string_error
    _option_executable_opts = _option_first_chunk_string_error
    _option_args = _option_first_chunk_string_error
    _option_jupyter_kernel = _option_first_chunk_string_error
    _option_jupyter_timeout = _option_first_chunk_int_error
    _option_save = _option_first_chunk_bool_error
    _option_save_as = _option_first_chunk_string_error
    _option_live_output = _option_first_chunk_bool_error

    _option_example = _option_bool_can_exec_error
    _option_lang = _option_str_error


    def _option_complete(self, key, value):
        if not isinstance(value, bool):
            self.code_chunk.errors.append(message.SourceError(f'Invalid "{key}" value "{value}"'))
        elif not self.code_chunk.execute:
            self.code_chunk.errors.append(message.SourceError('Option "complete" is only compatible with executed code chunks'))
        elif self.code_chunk.command == 'repl':
            self.code_chunk.errors.append(message.SourceError('Option "complete" is not compatible with "repl" command'))
        elif self.code_chunk.is_expr and not value:
            self.code_chunk.errors.append(message.SourceError('Option "complete" value "false" is incompatible with expr command'))
        elif self['outside_main']:
            # Technically, this is only required for complete=true, but
            # prohibiting it in all cases is more consistent
            self.code_chunk.errors.append(message.SourceError('Option "complete" is incompatible with "outside_main" value "true"'))
        else:
            self[key] = value


    def _option_copy(self, key, value):
        if not isinstance(value, str):
            self.code_chunk.errors.append(message.SourceError(f'Invalid "{key}" value "{value}"'))
        elif 'include' in self:
            self.code_chunk.errors.append(message.SourceError('Option "copy" is incompatible with "include"'))
        else:
            values = [x.strip() for x in value.split('+')]
            if not all(v.isidentifier() for v in values):
                invalid_values = ', '.join(f'"{v}"' for v in values if not v.isidentifier())
                self.code_chunk.errors.append(message.SourceError(f'Option "{key}" has invalid, non-identifier value(s) {invalid_values}'))
            elif self.get('name') in values:
                self.code_chunk.errors.append(message.SourceError('Code chunk cannot copy itself'))
            else:
                self[key] = values


    def _option_expand_tabs(self, key, value):
        if key == 'expand_tabs':
            key = 'code_expand_tabs'
        self._option_bool_can_exec_error(key, value)

    _option_markup_expand_tabs = _option_expand_tabs
    _option_copied_markup_expand_tabs = _option_expand_tabs
    _option_code_expand_tabs = _option_expand_tabs
    _option_stdout_expand_tabs = _option_expand_tabs
    _option_stderr_expand_tabs = _option_expand_tabs


    def _option_first_number(self, key, value):
        if not ((isinstance(value, int) and value > 0) or (isinstance(value, str) and value == 'next')):
            self.code_chunk.errors.append(message.CanExecSourceError(f'Invalid "{key}" value "{value}"'))
        else:
            if key == 'first_number':
                key = 'code_first_number'
            self[key] = value

    _option_markup_first_number = _option_first_number
    _option_copied_markup_first_number = _option_first_number
    _option_code_first_number = _option_first_number
    _option_stdout_first_number = _option_first_number
    _option_stderr_first_number = _option_first_number


    def _option_rewrap_lines(self, key, value):
        if key == 'rewrap_lines':
            key = 'code_rewrap_lines'
        self._option_bool_can_exec_error(key, value)

    _option_markup_rewrap_lines = _option_rewrap_lines
    _option_copied_markup_rewrap_lines = _option_rewrap_lines
    _option_code_rewrap_lines = _option_rewrap_lines
    _option_stdout_rewrap_lines = _option_rewrap_lines
    _option_stderr_rewrap_lines = _option_rewrap_lines


    def _option_rewrap_width(self, key, value):
        if key == 'rewrap_width':
            key = 'code_rewrap_width'
        self._option_positive_int_can_exec_error(key, value)

    _option_markup_rewrap_width = _option_rewrap_width
    _option_copied_markup_rewrap_width = _option_rewrap_width
    _option_code_rewrap_width = _option_rewrap_width
    _option_stdout_rewrap_width = _option_rewrap_width
    _option_stderr_rewrap_width = _option_rewrap_width


    def _option_hide(self, key, value,
                     display_values=set(['markup', 'copied_markup', 'code', 'stdout', 'stderr', 'expr', 'rich_output'])):
        # 'hide' may be processed during `finalize_after_copy()` to allow for
        # 'show' and `.is_expr` inheritance.
        if not isinstance(value, str):
            self.code_chunk.errors.append(message.CanExecSourceError(f'Invalid "{key}" value "{value}"'))
        elif 'show' in self.custom_options:
            # 'hide' checks for 'show' conflict, so 'show' does not.  Check
            # in `custom_options` since there's a default 'show' in `self`.
            self.code_chunk.errors.append(message.CanExecSourceError('Option "hide" cannot be used with "show"'))
        elif value == 'all':
            self['show'] = ODict()
        else:
            hide_values = value.replace(' ', '').split('+')
            if not all(v in display_values for v in hide_values):
                self.code_chunk.errors.append(message.CanExecSourceError(f'Invalid "{key}" value "{value}"'))
            else:
                for v in hide_values:
                    self['show'].pop(v, None)


    def _option_hide_markup_keys(self, key, value):
        if not isinstance(value, str):
            self.code_chunk.errors.append(message.CanExecSourceError(f'Invalid "{key}" value "{value}"'))
        else:
            # No need to check keys for validity; this is a display option.
            hide_keys = set(value.replace(' ', '').split('+'))
            hide_keys.add('hide_markup_keys')
            self[key] = hide_keys


    def _option_include(self, key, value):
        # Include() does its own value check, so this isn't technically needed
        if not isinstance(value, dict):
            self.code_chunk.errors.append(message.SourceError(f'Invalid "{key}" value "{value}"'))
        elif 'copy' in self:
            self.code_chunk.errors.append(message.SourceError('Option "include" is incompatible with "copy"'))
        else:
            include = Include(self.code_chunk, value)
            if include.code_lines:
                self[key] = include


    def _option_line_numbers(self, key, value):
        if not isinstance(value, bool):
            self.code_chunk.errors.append(message.CanExecSourceError(f'Invalid "{key}" value "{value}"'))
        else:
            if key == 'line_numbers':
                key = 'code_line_numbers'
            self[key] = value

    _option_markup_line_numbers = _option_line_numbers
    _option_copied_markup_line_numbers = _option_line_numbers
    _option_code_line_numbers = _option_line_numbers
    _option_stdout_line_numbers = _option_line_numbers
    _option_stderr_line_numbers = _option_line_numbers


    def _option_name(self, key, value):
        if not isinstance(value, str):
            self.code_chunk.errors.append(message.CanExecSourceError(f'Invalid "{key}" value "{value}"'))
        elif not value.isidentifier():
            self.code_chunk.errors.append(message.CanExecSourceError('Option "{0}" has invalid, non-identifier value "{1}"'.format(key, value)))
        elif 'copy' in self and key in self['copy']:
            self.code_chunk.errors.append(message.SourceError('Code chunk cannot copy itself'))
        else:
            self[key] = value


    def _option_outside_main(self, key, value):
        if not isinstance(value, bool):
            self.code_chunk.errors.append(message.SourceError(f'Invalid "{key}" value "{value}"'))
        elif not self.code_chunk.execute:
            self.code_chunk.errors.append(message.SourceError('Option "outside_main" is only compatible with executed code chunks'))
        elif self.code_chunk.command == 'repl':
            self.code_chunk.errors.append(message.SourceError('Option "outside_main" is not compatible with "repl" command'))
        elif self.code_chunk.is_expr and value:
            self.code_chunk.errors.append(message.SourceError('Option "outside_main" value "true" is incompatible with expr command'))
        elif value and 'complete' in self.custom_options:
            self.code_chunk.errors.append(message.SourceError('Option "outside_main" value "true" is incompatible with "complete"'))
        else:
            self['complete'] = None
            self[key] = value


    def _option_source(self, key, value):
        if not isinstance(value, str):
            self.code_chunk.errors.append(message.SourceError(f'Invalid "{key}" value "{value}"'))
        elif self.code_chunk.execute and self.code_chunk.command is not None:
            # Always preserve sources for unknown commands, so that these
            # sources can be marked as having potential errors later
            self.code_chunk.errors.append(message.SourceError('Option "source" is only compatible with non-executed code chunks; otherwise, use "session"'))
        elif not value.isidentifier():
            self.code_chunk.errors.append(message.SourceError('Option "{0}" has invalid, non-identifier value "{1}"'.format(key, value)))
        else:
            self[key] = value


    def _option_session(self, key, value):
        if not isinstance(value, str):
            self.code_chunk.errors.append(message.SourceError(f'Invalid "{key}" value "{value}"'))
        elif not self.code_chunk.execute and self.code_chunk.command is not None:
            # Always preserve sessions for unknown commands, so that these
            # sessions can be marked as having potential errors later
            self.code_chunk.errors.append(message.SourceError('Option "session" is only compatible with executed code chunks; otherwise, use "source"'))
        elif not value.isidentifier():
            self.code_chunk.errors.append(message.SourceError('Option "{0}" has invalid, non-identifier value "{1}"'.format(key, value)))
        else:
            self[key] = value


    mime_map = {'latex': 'text/latex',
                'html': 'text/html',
                'markdown': 'text/markdown',
                'plain': 'text/plain',
                'png': 'image/png',
                'jpg': 'image/jpeg',
                'jpeg': 'image/jpeg',
                'svg': 'image/svg+xml',
                'pdf': 'application/pdf'}

    mime_map_with_text_display = {}
    rich_text_default_display = {}
    for k, v in mime_map.items():
        mime_map_with_text_display[k] = v
        if v.startswith('text/'):
            mime_map_with_text_display[k+':raw'] = v
            mime_map_with_text_display[k+':verbatim'] = v
            mime_map_with_text_display[k+':verbatim_or_empty'] = v
            if k == 'plain':
                rich_text_default_display[k] = 'verbatim'
            else:
                rich_text_default_display[k] = 'raw'


    def _option_show(self, key, value):
        # 'show' may be processed during `finalize_after_copy()` to allow for
        # 'show' and `.is_expr` inheritance.  'hide' checks for 'show'
        # conflict, so 'show' does not.
        if not (isinstance(value, str) or value is None):
            self.code_chunk.errors.append(message.CanExecSourceError(f'Invalid "{key}" value "{value}"'))
        elif value in ('none', None):
            self[key] = ODict()
        else:
            value_processed = ODict()
            for output_and_format in value.replace(' ', '').split('+'):
                if ':' not in output_and_format:
                    output = output_and_format
                    format = None
                else:
                    output, format = output_and_format.split(':', 1)
                if output in value_processed:
                    self.code_chunk.errors.append(message.CanExecSourceError('Option "{0}" value "{1}" contains duplicate "{2}"'.format(key, value, output)))
                    continue
                if output in ('markup', 'copied_markup', 'code', 'repl'):
                    if format is None:
                        format = 'verbatim'
                    elif format != 'verbatim':
                        self.code_chunk.errors.append(message.CanExecSourceError('Invalid "{0}" sub-value "{1}"'.format(key, output_and_format)))
                        continue
                    if output == 'copied_markup' and 'copy' not in self.custom_options:
                        self.code_chunk.errors.append(message.CanExecSourceError('Invalid "{0}" sub-value "{1}"; can only be used with "copy"'.format(key, output_and_format)))
                        continue
                elif output in ('stdout', 'stderr'):
                    if format is None:
                        format = 'verbatim'
                    elif format not in ('verbatim', 'verbatim_or_empty', 'raw'):
                        self.code_chunk.errors.append(message.CanExecSourceError('Invalid "{0}" sub-value "{1}"'.format(key, output_and_format)))
                        continue
                elif output == 'expr':
                    if not self.code_chunk.is_expr:
                        self.code_chunk.errors.append(message.CanExecSourceError('Invalid "{0}" sub-value "{1}"'.format(key, output_and_format)))
                        continue
                    if format is None:
                        format = 'raw'
                    elif format not in ('verbatim', 'verbatim_or_empty', 'raw'):
                        self.code_chunk.errors.append(message.CanExecSourceError('Invalid "{0}" sub-value "{1}"'.format(key, output_and_format)))
                        continue
                elif output == 'rich_output':
                    if format is None:
                        format = self._default_rich_output
                    else:
                        format = format.split('|')
                        if not all(fmt in self.mime_map_with_text_display for fmt in format):
                            self.code_chunk.errors.append(message.CanExecSourceError('Invalid "{0}" sub-value "{1}"'.format(key, output_and_format)))
                            continue
                else:
                    self.code_chunk.errors.append(message.CanExecSourceError('Invalid "{0}" sub-value "{1}"'.format(key, output_and_format)))
                    continue
                value_processed[output] = format
            self[key] = value_processed


    def _option_tab_size(self, key, value):
        if key == 'tab_size':
            key = 'code_tab_size'
        self._option_positive_int_can_exec_error(key, value)

    _option_markup_tab_size = _option_tab_size
    _option_copied_markup_tab_size = _option_tab_size
    _option_code_tab_size = _option_tab_size
    _option_stdout_tab_size = _option_tab_size
    _option_stderr_tab_size = _option_tab_size




class CodeChunk(object):
    '''
    Base class for code chunks.
    '''
    def __init__(self,
                 command: str,
                 code: Union[str, List[str]],
                 custom_options: dict,
                 *,
                 origin_name: Optional[str]=None,
                 origin_start_line_number: Optional[int]=None,
                 inline: Optional[bool]=None):
        self.errors: message.ErrorMessageList
        self.warnings: message.WarningMessageList
        self.__pre_init__()

        self.key: Optional[CodeKey]  # Set during session/source creation

        if command not in self.commands:
            if command is None:
                self.errors.append(message.SourceError('Missing valid Codebraid command'))
            else:
                self.errors.append(message.SourceError('Unknown Codebraid command "{0}"'.format(command)))
            self.command = None
        else:
            self.command = command
        if command == 'expr' and not inline:
            self.errors.append(message.SourceError('Codebraid command "{0}" is only allowed inline'.format(command)))
        if command == 'repl' and inline:
            self.errors.append(message.SourceError('Codebraid command "{0}" is not supported inline'.format(command)))
        self.execute = self._default_execute[command]
        if command == 'expr' or (inline and command == 'nb'):
            self.is_expr = True
        else:
            # For the paste command, or code with 'copy', this value can
            # change later due to inheritance
            self.is_expr = False
        self.origin_name = origin_name
        self.origin_start_line_number = origin_start_line_number
        self.inline = inline

        # Check for len(code_lines) > 1 for inline later
        self._code_str = None
        self._placeholder_code_str = None
        if isinstance(code, list):
            code_lines = code
        else:
            code_lines = util.splitlines_lf(code, last_lf_stripped=True) or ['']
        if 'copy' not in custom_options and 'include' not in custom_options:
            if inline and len(code_lines) > 1:
                self.errors.append(message.SourceError('Inline code cannot be longer that 1 line'))
            self.code_lines = code_lines
            self.placeholder_code_lines = []
        else:
            if inline:
                if len(code_lines) > 1 or code_lines[0] not in ('', ' ', '_'):
                    self.errors.append(message.SourceError('Invalid placeholder code for copy or include (need space or underscore)'))
            elif len(code_lines) > 1 or code_lines[0].rstrip(' ') not in ('', '_'):
                self.errors.append(message.SourceError('Invalid placeholder code for copy or include (need empty, space, or underscore)'))
            # Copying or including code could result in more than one line of
            # code in an inline context.  That is only an issue if the code is
            # actually displayed.  This is checked later when code is
            # included/copied.
            self.placeholder_code_lines = code_lines
            self.code_lines = []

        self.options = Options(self, custom_options)

        if 'include' in self.options and not self.errors:
            # Copy over include only if no source errors -- otherwise it isn't
            # used and 'show' may not exist
            include = self.options['include']
            if inline and 'code' in self.options['show'] and len(include.code_lines) > 1:
                self.errors.append(message.SourceError('Cannot include and then display multiple lines of code in an inline context'))
            else:
                self.code_lines = include.code_lines

        if command == 'paste':
            if 'copy' not in custom_options:
                self.errors.append(message.SourceError('Command "paste" cannot be used without specifying a target via "copy"'))
            self.needs_to_copy = True
        else:
            self.needs_to_copy = False
        if 'copy' in self.options:
            self.copy_chunks = []

        self.code_collection: Optional[CodeCollection] = None
        self.index: Optional[int] = None
        self.output_index = None
        self.stdout_lines = []
        self.stderr_lines = []
        self.repl_lines = []
        self.rich_output = None
        self.expr_lines = []
        self.markup_start_line_number = 1
        self.code_start_line_number = 1
        self.stdout_start_line_number = 1
        self.stderr_start_line_number = 1

    @property
    def session(self):
        if self.execute:
            return self.code_collection
        raise TypeError

    @property
    def source(self):
        if not self.execute:
            return self.code_collection
        raise TypeError


    def __pre_init__(self):
        '''
        Create lists of errors and warnings.  Subclasses may need to record
        errors or warnings during preprocessing, before they are ready
        to call `super().__init__()`
        '''
        if not hasattr(self, 'errors'):
            self.errors = message.ErrorMessageList()
        elif not isinstance(self.errors, message.ErrorMessageList):
            raise TypeError
        if not hasattr(self, 'warnings'):
            self.warnings = message.WarningMessageList()
        elif not isinstance(self.warnings, message.WarningMessageList):
            raise TypeError


    commands = set(['code', 'expr', 'nb', 'paste', 'repl', 'run'])

    _default_execute = collections.defaultdict(lambda: False,  # Unknown command -> do not run
                                               {k: True for k in ('expr', 'nb', 'repl', 'run')})


    @property
    def code_str(self):
        code_str = self._code_str
        if code_str is not None:
            return code_str
        code_str = '\n'.join(self.code_lines)
        self._code_str = code_str
        return code_str

    @property
    def placeholder_code_str(self):
        placeholder_code_str = self._placeholder_code_str
        if placeholder_code_str is not None:
            return placeholder_code_str
        placeholder_code_str = '\n'.join(self.placeholder_code_lines)
        self._placeholder_code_str = placeholder_code_str
        return placeholder_code_str


    def finalize_line_numbers(self, code_start_line_number):
        # Line numbers can't be precalculated since they are determined by
        # how a session is assembled across potentially multiple sources
        self.code_start_line_number = code_start_line_number


    @property
    def attr_hash(self):
        raise NotImplementedError

    @property
    def code_hash(self):
        raise NotImplementedError

    def only_code_output(self, format):
        raise NotImplementedError


    def finalize_after_copy(self):
        '''
        Finalize options.  This can be redefined by subclasses so that they
        can modify themselves based on inherited 'lang' or 'show'.
        '''
        self.options.finalize_after_copy()


    def copy_code(self):
        '''
        Copy code for 'copy' option.  Code is copied before execution, which
        is more flexible.  Output (stdout, stderr, expr, etc.) must be copied
        separately after execution.

        This should only be invoked for a code chunk with no errors that would
        prevent execution, with copy targets that all exist and have no errors
        that would prevent execution.
        '''
        copy_chunks = self.copy_chunks
        if any(cc.is_expr for cc in copy_chunks):
            if len(copy_chunks) > 1:
                invalid_cc_names = ', '.join(cc.options['name'] for cc in copy_chunks if cc.is_expr)
                self.errors.append(message.SourceError('Cannot copy multiple code chunks when some are expressions: {0}'.format(invalid_cc_names)))
            if self.command in ('paste', 'code'):
                # Some commands inherit expression status.  The code command
                # inherits so that subsequent copying doesn't result in
                # incorrectly concatenated expressions.  Since the code
                # command never has output, this has no display side effects.
                self.is_expr = True
                self.expr_lines = []
            elif not self.is_expr:
                self.errors.append(message.SourceError('A non-expression command cannot copy an expression code chunk'))
        elif self.is_expr:
            self.errors.append(message.SourceError('An expression command cannot copy a non-expression code chunk'))
        if self.errors:
            return
        # Finalization must come after any potential `.is_expr` modifications
        self.finalize_after_copy()
        if self.inline and 'code' in self.options['show'] and (len(copy_chunks) > 1 or len(copy_chunks[0].code_lines) > 1):
            self.errors.append(message.SourceError('Cannot copy and then display multiple lines of code in an inline context'))
            return
        if len(copy_chunks) == 1:
            self.code_lines = copy_chunks[0].code_lines
        else:
            self.code_lines = [line for x in copy_chunks for line in x.code_lines]
        if self.command == 'paste':
            if all(cc.command == 'code' for cc in copy_chunks):
                # When possible, simplify the copying resolution process
                self.needs_to_copy = False


    def copy_output(self):
        '''
        Copy output (stdout, stderr, expr) for 'copy' option.  This must be
        copied separately from code, after execution.

        This should only be invoked for a code chunk with no source errors,
        with copy targets that all exist and have no source errors.
        '''
        if self.command != 'paste':
            raise TypeError
        copy_chunks = self.copy_chunks
        # The case of all code chunks being code commands has already been
        # handled in `copy_code()`
        if any(cc.command == 'paste' for cc in copy_chunks):
            if len(copy_chunks) > 1:
                if all(cc.command == 'paste' for cc in copy_chunks):
                    self.errors.append(message.SourceError('Can only copy a single paste code chunk; cannot combine multiple paste chunks'))
                else:
                    self.errors.append(message.SourceError('Cannot copy a mixture of paste and other code chunks'))
        elif any(cc.execute for cc in copy_chunks):
            if not all(cc.execute for cc in copy_chunks):
                self.errors.append(message.SourceError('Copying output of multiple code chunks requires that all or none are executed'))
            elif len(copy_chunks) > 1:
                if len(set(cc.session for cc in copy_chunks)) > 1:
                    self.errors.append(message.SourceError('Cannot copy output from code chunks in multiple sessions'))
                elif any(ccx.index != ccy.index-1 for ccx, ccy in zip(copy_chunks[:-1], copy_chunks[1:])):
                    if any(ccx is ccy for ccx, ccy in zip(copy_chunks[:-1], copy_chunks[1:])):
                        self.errors.append(message.SourceError('Cannot copy output of a single code chunk multiple times'))
                    elif any(ccx.index > ccy.index for ccx, ccy in zip(copy_chunks[:-1], copy_chunks[1:])):
                        self.errors.append(message.SourceError('Cannot copy output of code chunks out of order'))
                    else:
                        self.errors.append(message.SourceError('Cannot copy output of code chunks when some chunks in a sequence are omitted'))
        else:
            raise ValueError
        if self.errors:
            # If errors, discard what has already been copied
            self.code_lines = []
            return
        if len(copy_chunks) == 1:
            self.stdout_lines = copy_chunks[0].stdout_lines
            self.stderr_lines = copy_chunks[0].stderr_lines
            self.repl_lines = copy_chunks[0].repl_lines
            self.rich_output = copy_chunks[0].rich_output
        else:
            self.stdout_lines = [line for x in copy_chunks if x.stdout_lines for line in x.stdout_lines]
            self.stderr_lines = [line for x in copy_chunks if x.stderr_lines for line in x.stderr_lines]
            self.repl_lines = [line for x in copy_chunks if x.repl_lines for line in x.repl_lines]
            self.rich_output = [ro for x in copy_chunks if x.rich_output for ro in x.rich_output]
        if self.is_expr:
            # expr compatibility has already been checked in `copy_code()`
            self.expr_lines = copy_chunks[0].expr_lines
        self.stdout_start_line_number = copy_chunks[0].stdout_start_line_number
        self.stderr_start_line_number = copy_chunks[0].stderr_start_line_number
        self.needs_to_copy = False


    @property
    def as_markup_lines(self):
        raise NotImplementedError

    @property
    def as_example_markup_lines(self):
        raise NotImplementedError

    def layout_output(self, output_type, output_format, lines=None):
        '''
        Layout all forms of output, except for rich output that is not
        text/plain, by performing operations such as line rewrapping and tab
        expansion.  If `lines` is supplied, it is used.  Otherwise, the
        default lines (if any) are accessed for the specified output type.
        '''
        if lines is not None:
            if not lines and output_format == 'verbatim_or_empty':
                lines = ['\xa0']
            pass
        elif output_type == 'code':
            lines = self.code_lines
        elif output_type == 'repl':
            lines = self.repl_lines
        elif output_type in ('expr', 'stdout', 'stderr'):
            lines = getattr(self, output_type+'_lines')
            if not lines and output_format == 'verbatim_or_empty':
                lines = ['\xa0']
        elif output_type == 'markup':
            lines = self.as_markup_lines
        elif output_type == 'example_markup':
            lines = self.as_example_markup_lines
        elif output_type == 'copied_markup':
            if len(self.copy_chunks) == 1:
                lines = self.copy_chunks[0].as_markup_lines
            elif self.inline:
                lines = []
                for cc in self.copy_chunks:
                    lines.extend(cc.as_markup_lines)
            else:
                lines = []
                last_cc = self.copy_chunks[-1]
                for cc in self.copy_chunks:
                    lines.extend(cc.as_markup_lines)
                    if cc is not last_cc:
                        lines.append('')
        else:
            raise ValueError

        rewrap_lines = self.options.get(output_type+'_rewrap_lines', False)
        rewrap_width = self.options.get(output_type+'_rewrap_width', 78)
        expand_tabs = self.options.get(output_type+'_expand_tabs', False)
        tab_size = self.options.get(output_type+'_tab_size', 8)
        # This should be rewritten once rewrapping design is finalized, since
        # textwrap doesn't necessarily do everything as might be desired, and
        # the use of textwrap could be optimized if it continues to be used.
        # Nothing is done yet with tabs.
        if rewrap_lines:
            new_lines = []
            for line in lines:
                if not line:
                    new_lines.append(line)
                    continue
                line_stripped = line.lstrip(' \t')
                indent = line[:len(line)-len(line_stripped)]
                new_lines.extend(textwrap.wrap(line_stripped, width=rewrap_width-len(indent), initial_indent=indent, subsequent_indent=indent))
            lines = new_lines
        if self.inline:
            return ' '.join(lines)
        return '\n'.join(lines)
